import asyncio
import asyncio
import attrs
import inspect
import logging
import numpy as np
import pyvista as pv
from scipy.stats import trim_mean
from scipy.optimize import least_squares
from qtpy import QtWidgets, QtGui, QtCore
import typing as tp

from NaviNIBS.Navigator.GUI.EditWindows.ToolCalibrationWindow import ToolCalibrationWindow
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.Transforms import invertTransform, concatenateTransforms, applyTransform
from NaviNIBS.util.pyvista import Actor, setActorUserTransform, concatenateLineSegments
from NaviNIBS.util.pyvista import DefaultBackgroundPlotter, RemotePlotterProxy


logger = logging.getLogger(__name__)


@attrs.define(kw_only=True)
class VisualizedOrientation:
    """
    Note: this doesn't connect to any change signals from underlying sample, instead assuming that caller will
    re-instantiate for any changes as needed
    """
    _transf: np.ndarray
    _plotter: pv.Plotter
    _colorDepthIndicator: str = '#ba55d3'
    _colorHandleIndicator: str = '#9360db'
    _colorEndpointIndicator: str = '#7340bb'
    _opacity: float = 1.
    _lineWidth: float = 3.
    _style: str = 'linesPlusEndpoint'
    _actorKeyPrefix: str

    _actors: tp.Dict[str, Actor] = attrs.field(init=False, factory=dict)

    def __attrs_post_init__(self):
        match self._style:
            case 'linesPlusEndpoint':
                zOffset = 20
                handleLength = 2
                depthLine = pv.lines_from_points(np.asarray([[0, 0, 0], [0, 0, zOffset]]))
                handleLine = pv.lines_from_points(np.asarray([[0, 0, zOffset], [0, -handleLength, zOffset]]))
                actorKey = self._actorKeyPrefix + 'depthLine'
                self._actors[actorKey] = self._plotter.addLineSegments(
                                                         depthLine,
                                                         name=actorKey,
                                                         color=self._colorDepthIndicator,
                                                         width=self._lineWidth,
                                                         opacity=self._opacity)

                actorKey = self._actorKeyPrefix + 'handleLine'
                self._actors[actorKey] = self._plotter.addLineSegments(
                                                         handleLine,
                                                         name=actorKey,
                                                         color=self._colorHandleIndicator,
                                                         width=self._lineWidth,
                                                         opacity=self._opacity)

                actorKey = self._actorKeyPrefix + 'endpoint'
                self._actors[actorKey] = self._plotter.add_points(
                    np.asarray([[0, 0, 0]]),
                    color=self._colorEndpointIndicator,
                    name=actorKey,
                    opacity=self._opacity,
                    reset_camera=False,
                    point_size=8.
                )

            case _:
                raise NotImplementedError(f'Unexpected style: {self._style}')

        for actor in self._actors.values():
            setActorUserTransform(actor, self._transf)

        self._plotter.render()

    @property
    def actors(self):
        return self._actors

    @property
    def transf(self):
        return self._transf

    @transf.setter
    def transf(self, newTransf: np.ndarray):
        self._transf = newTransf
        for actor in self._actors.values():
            setActorUserTransform(actor, self._transf)


@attrs.define
class PointerCalibrationWindow(ToolCalibrationWindow):

    _samples: list[np.ndarray] = attrs.field(factory=list)  # list of pointer poses with common endpoint location
    """
    Note that these are orientations BEFORE applying any previous toolToTrackerTransf, even though visualization
    will include this previous toolToTrackerTransf
    """

    _pendingNewTransf: tp.Optional[np.ndarray] = attrs.field(init=False, default=None)  # result of calibration, not yet saved to tool

    _plotter: DefaultBackgroundPlotter = attrs.field(init=False)
    _liveVisual: tp.Optional[VisualizedOrientation] = attrs.field(init=False, default=None)
    _sampleVisuals_orig: list[VisualizedOrientation] = attrs.field(init=False, factory=list)
    _sampleVisuals_pending: list[VisualizedOrientation] = attrs.field(init=False, factory=list)

    _visualColors_orig: tuple[str, str, str] = ('#ba55d3', '#9360db', '#7340bb')
    _visualColors_pending: tuple[str, str, str] = ('#bad355', '#93db60', '#73bb40')
    _visualColors_live: tuple[str, str, str] = ('#d3ba55', '#db9360', '#bb7340')

    _instructions: QtWidgets.QTextEdit = attrs.field(init=False)
    _deleteLastSampleBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _acquireSampleBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _calibrateBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _revertBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _showOrigVisuals: QtWidgets.QCheckBox = attrs.field(init=False)
    _showPendingVisuals: QtWidgets.QCheckBox = attrs.field(init=False)
    _showLiveVisual: QtWidgets.QCheckBox = attrs.field(init=False)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._wdgt.setLayout(QtWidgets.QVBoxLayout())

        splitter = QtWidgets.QSplitter()
        self._wdgt.layout().addWidget(splitter)

        self._instructions = QtWidgets.QTextEdit('')
        self._instructions.setReadOnly(True)
        self._instructions.setMarkdown(inspect.cleandoc("""
            # Pointer endpoint calibration
            ## Instructions
            1. Choose a rigid, fixed reference point on a solid object in the room. The camera should not move relative to this fixed reference point throughout calibration.
            2. Position the pointer to be visible to the camera, with its endpoint precisely placed on the reference point.
            3. Press "Acquire sample" to record the stylus pose while the stylus tip is touching the reference point.
            4. Rotate the stylus while holding the endpoint fixed at the reference point, and then acquire another sample.
            5. Repeat rotate + acquire process for a number of samples. 
                - The more samples the better, as long as you maintain consistent contact with the reference point whenever sampling.
                - Try to get a broad range of angles sampled, including pivoting the stylus along ~150 degrees along the left-right axis, ~150 degrees along the front-back axis, and ~180 degrees rotating around the stylus shaft.
            6. When done acquiring samples, click "Process samples" to estimate a new endpoint.
            7. Double check that the samples with new transform show endpoints tightly clustered around a single point. If you accidentally deviated from the reference point during a small number of samples, those should be visible as outliers here.
            8. Close this window to accept the calibration.
        """))
        splitter.addWidget(self._instructions)

        splitRight = QtWidgets.QWidget()
        splitRight.setLayout(QtWidgets.QVBoxLayout())
        splitter.addWidget(splitRight)

        self._plotter = DefaultBackgroundPlotter()
        splitRight.layout().addWidget(self._plotter)

        if len(self._samples) > 0:
            raise NotImplementedError  # TODO: initialize corresponding sample actors

        # TODO: initialize actor to visualize live position of pointer

        btnContainer = QtWidgets.QWidget()
        btnCntLayout = QtWidgets.QGridLayout()
        btnContainer.setLayout(btnCntLayout)
        splitRight.layout().layout().addWidget(btnContainer)

        btn = QtWidgets.QPushButton('Delete last sample')
        btn.clicked.connect(lambda checked: self._onDeleteLastSampleClicked())
        btn.setEnabled(len(self._samples) > 0)
        btnCntLayout.addWidget(btn, 1, 0)
        self._deleteLastSampleBtn = btn

        btn = QtWidgets.QPushButton('Acquire sample')
        btn.clicked.connect(lambda checked: self._onAcquireSampleClicked())
        btn.setEnabled(False)
        btnCntLayout.addWidget(btn, 0, 0)
        self._acquireSampleBtn = btn

        btn = QtWidgets.QPushButton('Process samples')
        btn.clicked.connect(lambda checked: self._onCalibrateClicked())
        btn.setEnabled(len(self._samples) > 2)
        btnCntLayout.addWidget(btn, 0, 1)
        self._calibrateBtn = btn

        btn = QtWidgets.QPushButton('Revert calibration')
        btn.clicked.connect(lambda checked: self._onRevertClicked())
        btn.setEnabled(self._pendingNewTransf is not None)
        btnCntLayout.addWidget(btn, 1, 1)
        self._revertBtn = btn

        groupBox = QtWidgets.QGroupBox('Render')
        groupBox.setLayout(QtWidgets.QVBoxLayout())
        splitRight.layout().addWidget(groupBox)

        cb = QtWidgets.QCheckBox('Samples with original transform')
        cb.setStyleSheet(f'QCheckBox {{color: {self._visualColors_orig[2]} }}')
        cb.setChecked(True)
        cb.stateChanged.connect(self._onShowOrigVisualsChanged)
        groupBox.layout().addWidget(cb)
        self._showOrigVisuals = cb

        cb = QtWidgets.QCheckBox('Samples with new transform')
        cb.setStyleSheet(f'QCheckBox {{color: {self._visualColors_pending[2]} }}')
        cb.setChecked(True)
        cb.setEnabled(self._pendingNewTransf is not None)
        cb.stateChanged.connect(self._onShowPendingVisualsChanged)
        groupBox.layout().addWidget(cb)
        self._showPendingVisuals = cb

        cb = QtWidgets.QCheckBox('Live position')
        cb.setStyleSheet(f'QCheckBox {{color: {self._visualColors_live[2]} }}')
        cb.setChecked(True)
        cb.stateChanged.connect(self._onShowLiveVisualChanged)
        groupBox.layout().addWidget(cb)
        self._showLiveVisual = cb

        self.sigFinished.connect(self._onDialogFinished)

        self._wdgt.resize(QtCore.QSize(1000, 800))

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._finishInitialization_async))

    async def _finishInitialization_async(self):
        for btn in (
            self._deleteLastSampleBtn,
            self._acquireSampleBtn,
            self._calibrateBtn,
            self._revertBtn,
        ):
            btn.setEnabled(False)

        if isinstance(self._plotter, RemotePlotterProxy):
            await self._plotter.isReadyEvent.wait()

        self._deleteLastSampleBtn.setEnabled(len(self._samples) > 0)
        self._calibrateBtn.setEnabled(len(self._samples) > 2)
        self._revertBtn.setEnabled(self._pendingNewTransf is not None)

        self._onLatestPositionsChanged()

    def _onDeleteLastSampleClicked(self):
        if len(self._samples) > 0:
            logger.info(f'Deleting sample {len(self._samples)}')

            assert len(self._samples) == len(self._sampleVisuals_orig)
            if self._pendingNewTransf is not None:
                assert len(self._samples) == len(self._sampleVisuals_pending)

            self._samples.pop()

            visual = self._sampleVisuals_orig.pop()
            for actor in visual.actors.values():
                self._plotter.remove_actor(actor)

            if self._pendingNewTransf is not None:
                visual = self._sampleVisuals_pending.pop()
                for actor in visual.actors.values():
                    self._plotter.remove_actor(actor)

        if len(self._samples) == 0:
            self._deleteLastSampleBtn.setEnabled(False)

        if len(self._samples) < 3:
            self._calibrateBtn.setEnabled(False)

    def _onAcquireSampleClicked(self):
        toolTrackerToCameraTransf = self._positionsClient.getLatestTransf(self.toolToCalibrate.trackerKey)
        self._samples.append(toolTrackerToCameraTransf)

        # draw new visual for sample
        self._sampleVisuals_orig.append(VisualizedOrientation(
            transf=concatenateTransforms([
                self.toolToCalibrate.toolToTrackerTransf,
                toolTrackerToCameraTransf]),
            colorDepthIndicator=self._visualColors_orig[0],
            colorHandleIndicator=self._visualColors_orig[1],
            colorEndpointIndicator=self._visualColors_orig[2],
            plotter=self._plotter,
            actorKeyPrefix=f'OrigSample{len(self._samples)}'
        ))

        if not self._showOrigVisuals.isChecked():
            for actor in self._sampleVisuals_orig[-1].actors.values():
                actor.SetVisibility(False)

        if self._pendingNewTransf is not None:
            self._sampleVisuals_pending.append(VisualizedOrientation(
                transf=concatenateTransforms([
                    self._pendingNewTransf,
                    toolTrackerToCameraTransf]),
                colorDepthIndicator=self._visualColors_pending[0],
                colorHandleIndicator=self._visualColors_pending[1],
                colorEndpointIndicator=self._visualColors_pending[2],
                plotter=self._plotter,
                actorKeyPrefix=f'PendingSample{len(self._samples)}'
            ))
            if not self._showPendingVisuals.isChecked():
                for actor in self._sampleVisuals_pending[-1].actors.values():
                    actor.SetVisibility(False)

        self._deleteLastSampleBtn.setEnabled(True)
        if len(self._samples) > 2:
            self._calibrateBtn.setEnabled(True)

        self._plotter.reset_camera()

    def _onCalibrateClicked(self):

        # set up optimization problem
        def costFn(offset: np.ndarray) -> float:
            uncalibratedToCalibratedTransf = self.toolToCalibrate.toolToTrackerTransf.copy()
            uncalibratedToCalibratedTransf[0:3, -1] = offset

            sampledEndpoints = np.full((len(self._samples), 3), np.nan)
            for iS, sample in enumerate(self._samples):
                sampledEndpoints[iS, :] = applyTransform([uncalibratedToCalibratedTransf, sample], np.asarray([0, 0, 0]), doCheck=False)

            meanEndpoint = trim_mean(sampledEndpoints, 0.1, axis=0)

            dists = np.linalg.norm(sampledEndpoints - meanEndpoint, axis=1)**2

            return np.percentile(dists, 80)

        initialOffset = self.toolToCalibrate.toolToTrackerTransf[0:3, -1].T

        maxOffset = 500  # in mm
        bounds = (np.ones((3,))*-maxOffset/2, np.ones((3,))*maxOffset/2)

        logger.debug('Running endpoint calibration optimization')
        result = least_squares(
            fun=costFn,
            x0=initialOffset,
            bounds=bounds
            )
        logger.info(f'Optimization results: {result}')
        optimizedOffset = result.x

        if self._pendingNewTransf is not None:
            # destroy any visuals associated with previous pending transform
            for visual in self._sampleVisuals_pending:
                for actor in visual.actors.values():
                    self._plotter.remove_actor(actor)
            self._sampleVisuals_pending.clear()

        self._pendingNewTransf = self.toolToCalibrate.toolToTrackerTransf.copy()
        self._pendingNewTransf[0:3, -1] = optimizedOffset

        self._revertBtn.setEnabled(True)

        assert len(self._sampleVisuals_pending) == 0
        for iS, sample in enumerate(self._samples):
            toolTrackerToCameraTransf = sample
            self._sampleVisuals_pending.append(VisualizedOrientation(
                transf=concatenateTransforms([
                    self._pendingNewTransf,
                    toolTrackerToCameraTransf]),
                colorDepthIndicator=self._visualColors_pending[0],
                colorHandleIndicator=self._visualColors_pending[1],
                colorEndpointIndicator=self._visualColors_pending[2],
                plotter=self._plotter,
                actorKeyPrefix=f'PendingSample{iS}'
            ))

            if not self._showPendingVisuals.isChecked():
                for actor in self._sampleVisuals_pending[-1].actors.values():
                    actor.SetVisibility(False)

        self._plotter.render()

        self._showPendingVisuals.setEnabled(True)

        self._plotter.reset_camera()

    def _onRevertClicked(self):

        if self._pendingNewTransf is not None:
            for visual in self._sampleVisuals_pending:
                for actor in visual.actors.values():
                    self._plotter.remove_actor(actor)
            self._sampleVisuals_pending.clear()

        self._pendingNewTransf = None
        self._revertBtn.setEnabled(False)

    def _onShowOrigVisualsChanged(self, state: int):
        doShow = state > 0
        for visual in self._sampleVisuals_orig:
            for actor in visual.actors.values():
                actor.SetVisibility(doShow)

    def _onShowPendingVisualsChanged(self, state: int):
        doShow = state > 0
        for visual in self._sampleVisuals_pending:
            for actor in visual.actors.values():
                actor.SetVisibility(doShow)

    def _onShowLiveVisualChanged(self, state: int):
        doShow = state > 0
        if self._liveVisual is not None:
            for actor in self._liveVisual.actors.values():
                actor.SetVisibility(doShow)

    def _onDialogFinished(self, wasAccepted: bool):
        if self._pendingNewTransf is not None:
            self._session.tools[self._toolKeyToCalibrate].toolToTrackerTransf = self._pendingNewTransf
            logger.info('Saved {} calibration: {}'.format(self._toolKeyToCalibrate, self.toolToCalibrate.toolToTrackerTransf))

        self._plotter.close()

    def _onLatestPositionsChanged(self):
        if isinstance(self._plotter, RemotePlotterProxy) and not self._plotter.isReadyEvent.is_set():
            # plotter not ready yet
            return

        super()._onLatestPositionsChanged()
        if self.toolToCalibrate.isActive and self._positionsClient.getLatestTransf(self.toolToCalibrate.trackerKey, None) is not None:
            self._acquireSampleBtn.setEnabled(True)

            if self._pendingNewTransf is None:
                transfToolToTracker = self.toolToCalibrate.toolToTrackerTransf
            else:
                transfToolToTracker = self._pendingNewTransf
            transfToolTrackerToCamera = self._positionsClient.getLatestTransf(self.toolToCalibrate.trackerKey, None)
            transfToolToCamera = concatenateTransforms([
                        transfToolToTracker,
                        transfToolTrackerToCamera])
            if self._liveVisual is None:
                self._liveVisual = VisualizedOrientation(
                    transf=transfToolToCamera,
                    colorDepthIndicator=self._visualColors_live[0],
                    colorHandleIndicator=self._visualColors_live[1],
                    colorEndpointIndicator=self._visualColors_live[2],
                    plotter=self._plotter,
                    actorKeyPrefix=f'LivePose'
                )
            else:
                self._liveVisual.transf = transfToolToCamera
                for actor in self._liveVisual.actors.values():
                    actor.SetVisibility(self._showLiveVisual.isChecked())
            self._plotter.render()
        else:
            self._acquireSampleBtn.setEnabled(False)
            if self._liveVisual is not None:
                for actor in self._liveVisual.actors.values():
                    actor.SetVisibility(False)
                self._plotter.render()




import attrs
import logging
from qtpy import QtWidgets, QtCore, QtGui

from NaviNIBS.util.GUI.StyleSheets import setStyleSheetForInstanceOnly

logger = logging.getLogger(__name__)


class VerticalScrollArea(QtWidgets.QScrollArea):
    """
    Address issue with behavior of default ScrollArea when never showing horizantal scrollbar but allowing resize: child widgets can be truncated left/right.

    This class overrides the resizeEvent to ensure that the minimum width of the scroll area is always wide enough to show full width of child widgets.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)

    def resizeEvent(self, event: QtGui.QResizeEvent):
        width = self.widget().minimumSizeHint().width()
        if self.verticalScrollBar().isVisible():
            width += self.verticalScrollBar().width()
        self.setMinimumWidth(width)
        super().resizeEvent(event)

    def sizeHint(self) -> QtCore.QSize:
        """
        Modify size hint behavior to prevent scrollbar showing in some situations where the widget could be given more room to grow
        """
        sz = self.widget().sizeHint()
        sz.setHeight(sz.height()+10)
        return sz


@attrs.define
class QScrollContainer:
    _allowVerticalScrolling: bool = attrs.field(default=True)
    _allowHorizontalScrolling: bool = attrs.field(default=False)

    _innerContainerLayout: QtWidgets.QLayout = attrs.field(factory=QtWidgets.QVBoxLayout)
    _innerContainer: QtWidgets.QWidget = attrs.field(factory=QtWidgets.QWidget)
    _scrollArea: QtWidgets.QScrollArea = attrs.field(init=False)

    def __attrs_post_init__(self):
        self._innerContainer.setLayout(self._innerContainerLayout)

        if self._allowVerticalScrolling and not self._allowHorizontalScrolling:
            self._scrollArea = VerticalScrollArea()
        else:
            self._scrollArea = QtWidgets.QScrollArea()

            self._scrollArea.setHorizontalScrollBarPolicy(
                QtCore.Qt.ScrollBarAsNeeded if self._allowHorizontalScrolling
                else QtCore.Qt.ScrollBarAlwaysOff)
            self._scrollArea.setVerticalScrollBarPolicy(
                QtCore.Qt.ScrollBarAsNeeded if self._allowVerticalScrolling
                else QtCore.Qt.ScrollBarAlwaysOff)

        setStyleSheetForInstanceOnly(self._scrollArea, 'background: transparent;')

        self._scrollArea.setWidgetResizable(True)

        self._scrollArea.setWidget(self._innerContainer)  # must happen after innerContainer's layout is set
        setStyleSheetForInstanceOnly(self._innerContainer, 'background-color: transparent;')

    @property
    def scrollArea(self):
        return self._scrollArea

    @property
    def innerContainer(self):
        return self._innerContainer

    @property
    def innerContainerLayout(self):
        return self._innerContainer.layout()


from __future__ import annotations

import asyncio

import attrs
from datetime import datetime
import logging
from math import ceil
import numpy as np
import os
import pathlib
import pyvista as pv
import qtawesome as qta
from qtpy import QtWidgets, QtGui, QtCore
import shutil
import typing as tp

from . import MainViewPanel
from NaviNIBS.Navigator.GUI.Widgets.MRIViews import MRISliceView
from NaviNIBS.Navigator.GUI.Widgets.SurfViews import SurfSliceView, Surf3DView
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.GUI.Icons import getIcon
from NaviNIBS.util.GUI.QFileSelectWidget import QFileSelectWidget
from NaviNIBS.Navigator.Model.Session import Session


logger = logging.getLogger(__name__)


@attrs.define()
class HeadModelPanel(MainViewPanel):
    _key: str = 'Set head model'
    _icon: QtGui.QIcon = attrs.field(init=False, factory=lambda: getIcon('mdi6.head-cog-outline'))
    _filepathWdgt: QFileSelectWidget = attrs.field(init=False)
    _activeSurfWidget: QtWidgets.QListWidget = attrs.field(init=False)
    _views: tp.Dict[str, tp.Union[SurfSliceView, Surf3DView]] = attrs.field(init=False, factory=dict)

    finishedAsyncInit: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def canBeEnabled(self) -> tuple[bool, str | None]:
        if self.session is None:
            return False, 'No session set'
        if not self.session.MRI.isSet:
            return False, 'No MRI set'
        return True, None

    def _finishInitialization(self):
        # don't initialize computationally-demanding views until panel is activated (viewed)

        super()._finishInitialization()

        self._wdgt.setLayout(QtWidgets.QVBoxLayout())

        wdgt = QFileSelectWidget(browseMode='getOpenFilename',
                                 extFilters='Gmsh (*.msh)')
        wdgt.sigFilepathChanged.connect(self._onBrowsedNewFilepath)
        self._wdgt.layout().addWidget(wdgt)
        self._filepathWdgt = wdgt

        containerWdgt = QtWidgets.QWidget()
        containerWdgt.setLayout(QtWidgets.QFormLayout())
        self._activeSurfWidget = QtWidgets.QListWidget()
        self._activeSurfWidget.itemSelectionChanged.connect(lambda *args, **kwargs: self._onSurfSelectionChanged())
        #self._activeSurfWidget.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Maximum)
        containerWdgt.layout().addRow('Surfaces', self._activeSurfWidget)
        containerWdgt.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Maximum)
        self._wdgt.layout().addWidget(containerWdgt)

        containerWdgt = QtWidgets.QWidget()
        containerWdgt.setSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        containerLayout = QtWidgets.QGridLayout()
        containerWdgt.setLayout(containerLayout)
        self._wdgt.layout().addWidget(containerWdgt)
        for iRow, iCol, key in ((0, 1, 'x'), (0, 0, 'y'), (1, 0, 'z'), (1, 1, '3D')):
            if key in ('x', 'y', 'z'):
                self._views[key] = SurfSliceView(normal=key)
            elif key == '3D':
                self._views[key] = Surf3DView(normal=key, surfOpacity=1)
            else:
                raise NotImplementedError()

            self._views[key].sigSliceOriginChanged.connect(lambda key=key: self._onSliceOriginChanged(sourceKey=key))

            containerLayout.addWidget(self._views[key].wdgt, iRow, iCol)

        if self.session is not None:
            self._onPanelInitializedAndSessionSet()

    def _onSliceOriginChanged(self, sourceKey: str):
        for key, view in self._views.items():
            if key == sourceKey:
                continue
            view.sliceOrigin = self._views[sourceKey].sliceOrigin

    def _onSessionSet(self):
        super()._onSessionSet()

        if self._hasInitialized:
            self._onPanelInitializedAndSessionSet()

    def _onPanelInitializedAndSessionSet(self):
        self._updateFilepath()
        self._updateRelativeToPath()
        self.session.sigInfoChanged.connect(self._onSessionInfoChanged)
        self.session.headModel.sigFilepathChanged.connect(self._updateFilepath)
        self.session.headModel.sigDataChanged.connect(self._onHeadModelUpdated)

        for key, view in self._views.items():
            view.session = self.session

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._finishInitialization_async))

    async def _finishInitialization_async(self):
        for view in self._views.values():
            await view.finishedAsyncInit.wait()

        self.finishedAsyncInit.set()

    def _onHeadModelUpdated(self, whatChanged: str):
        prevSelected = self._activeSurfWidget.selectedItems()
        if len(prevSelected) > 0:
            prevSelectedKey = prevSelected[0].text()
        else:
            prevSelectedKey = None
        self._activeSurfWidget.clear()
        if self.session is not None and self.session.headModel.isSet:
            for key in self.session.headModel.surfKeys:
                self._activeSurfWidget.addItem(key)
            if prevSelectedKey is not None:
                selectKey = prevSelectedKey
            else:
                selectKey = 'gmSurf'
            self._activeSurfWidget.setCurrentItem(self._activeSurfWidget.findItems(selectKey, QtCore.Qt.MatchExactly)[0])

        self._activeSurfWidget.setMaximumHeight(ceil(self._activeSurfWidget.sizeHintForRow(0) * (self._activeSurfWidget.count() + 0.2)))

        for key, view in self._views.items():
            view.updateView()

    def _onSurfSelectionChanged(self):
        selected = self._activeSurfWidget.selectedItems()
        if len(selected) == 0:
            return
        selectedKey = selected[0].text()
        for key, view in self._views.items():
            view.activeSurf = selectedKey

    def _updateFilepath(self):
        self._filepathWdgt.filepath = self.session.headModel.filepath

    def _onSessionInfoChanged(self, whatChanged: tp.Optional[list[str]] = None):
        if whatChanged is None or 'filepath' in whatChanged:
            self._updateRelativeToPath()

    def _updateRelativeToPath(self):
        self._filepathWdgt.showRelativeTo = os.path.dirname(self.session.filepath)
        self._filepathWdgt.showRelativePrefix = '<session>'

    def _onBrowsedNewFilepath(self, newFilepath: str):
        self.session.headModel.filepath = newFilepath


import logging
from qtpy import QtWidgets, QtCore, QtGui


logger = logging.getLogger(__name__)


class QLineEditWithValidationFeedback(QtWidgets.QLineEdit):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.textEdited.connect(self._onValidityChanged)

    def _onValidityChanged(self, newText):
        validity, _, _ = self.validator().validate(newText, self.cursorPosition())
        logger.debug('checking validity: %s' % validity)
        if validity == QtGui.QValidator.State.Acceptable:
            self.setStyleSheet(
                'QLineEdit, QPlainTextEdit { font-weight : normal; background-color : white; }')
        elif validity == QtGui.QValidator.State.Intermediate:
            self.setStyleSheet(
                'QLineEdit, QPlainTextEdit { font-weight : normal; background-color : red; }')
        elif validity == QtGui.QValidator.State.Invalid:
            self.setStyleSheet(
                'QLineEdit, QPlainTextEdit { font-weight : normal; background-color : red; }')
        else:
            raise NotImplementedError()


import attrs
import numpy as np
import typing as tp

from NaviNIBS.util.numpy import attrsWithNumpyAsDict, attrsWithNumpyFromDict

positionsServerHostname = '127.0.0.1'
positionsServerPubPort = 18950
positionsServerCmdPort = 18951


@attrs.define
class TimestampedToolPosition:
    time: float
    transf: tp.Optional[np.ndarray]  # None indicates invalid / not tracked
    relativeTo: str = 'world'

    def asDict(self) -> tp.Dict[str, tp.Any]:
        d = attrsWithNumpyAsDict(self, npFields=('transf',))
        return d

    @classmethod
    def fromDict(cls, d: tp.Dict[str, tp.Any]):
        return attrsWithNumpyFromDict(cls, d, npFields=('transf',))




from __future__ import annotations

import shutil

import attrs
from datetime import datetime
import nibabel as nib
import json
import logging
import numpy as np
import os
import pandas as pd
import pyvista as pv
import tempfile
import typing as tp
from typing import ClassVar

from NaviNIBS.util.attrs import attrsAsDict
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.numpy import array_equalish


logger = logging.getLogger(__name__)


SurfMesh = pv.PolyData
VolMesh = pv.PolyData


@attrs.define()
class HeadModel:
    _filepath: tp.Optional[str] = None
    """ 
    Path to .msh file in simnibs folder.
    (note that .msh file and other nested files in same parent dir will be used)
    """
    _skinSurf: tp.Optional[SurfMesh] = attrs.field(init=False, default=None)
    _csfSurf: tp.Optional[SurfMesh] = attrs.field(init=False, default=None)
    _gmSurf: tp.Optional[SurfMesh] = attrs.field(init=False, default=None)
    _skinSimpleSurf: tp.Optional[SurfMesh] = attrs.field(init=False, default=None)
    _gmSimpleSurf: tp.Optional[SurfMesh] = attrs.field(init=False, default=None)
    _eegPositions: tp.Optional[pd.DataFrame] = attrs.field(init=False, default=None)

    sigFilepathChanged: Signal = attrs.field(init=False, factory=Signal)
    sigDataChanged: Signal = attrs.field(init=False, factory=lambda: Signal((tp.Optional[str],)))
    """
    emits key `which` indicating what changed, e.g. which='gmSurf'; 
    if None all should be assumed to have changed
    """

    def __attrs_post_init__(self):
        self.sigFilepathChanged.connect(self._onFilepathChanged)
        self.validateFilepath(self._filepath)

    @property
    def _m2mDir(self):
        parentDir = os.path.dirname(self.filepath)  # simnibs results dir
        subStr = os.path.splitext(os.path.basename(self.filepath))[0]  # e.g. 'sub-1234'
        m2mDir = os.path.join(parentDir, 'm2m_' + subStr)
        assert os.path.exists(m2mDir), 'm2m folder not found. Are full SimNIBS results available next to specified .msh file?'
        return m2mDir

    @property
    def skinSurfPath(self):
        return os.path.join(self._m2mDir, 'skin.stl')

    @property
    def csfSurfPath(self):
        return os.path.join(self._m2mDir, 'csf.stl')

    @property
    def gmSurfPath(self):
        return os.path.join(self._m2mDir, 'gm.stl')

    def loadCache(self, which: str):
        if not self.isSet:
            logger.warning('Load data requested, but no filepath set. Returning.')
            return

        m2mDir = self._m2mDir

        if which in ('skinSurf', 'csfSurf', 'gmSurf'):
            if which == 'gmSurf':
                meshPath = self.gmSurfPath
            elif which == 'csfSurf':
                meshPath = self.csfSurfPath
            elif which == 'skinSurf':
                meshPath = self.skinSurfPath
            else:
                raise NotImplementedError()

            logger.info('Loading {} mesh from {}'.format(which, meshPath))
            mesh = pv.read(meshPath)

            setattr(self, '_' + which, mesh)

        elif which in ('skinSimpleSurf', 'gmSimpleSurf'):
            if which == 'gmSimpleSurf':
                mesh = self.gmSurf
            elif which == 'skinSimpleSurf':
                mesh = self.skinSurf
            else:
                raise NotImplementedError

            if True:  # TODO: debug, set to True
                logger.info(f'Simplifying mesh for {which}')
                mesh = mesh.decimate(0.8)
                logger.debug('Done simplifying mesh')

            setattr(self, '_' + which, mesh)

        elif which == 'eegPositions':
            csvPath = os.path.join(m2mDir, 'eeg_positions', 'EEG10-10_UI_Jurak_2007.csv')
            columnLabels = ('type', 'x', 'y', 'z', 'label')
            logger.info('Loading EEG positions from {}'.format(csvPath))
            self._eegPositions = pd.read_csv(csvPath, names=columnLabels, index_col='label')
            assert self._eegPositions.shape[1] == len(columnLabels) - 1

        else:
            raise NotImplementedError()

        self.sigDataChanged.emit(which)

    def clearCache(self, which: str):

        if which == 'all':
            allKeys = ('skinSurf', 'csfSurf', 'gmSurf', 'skinSimpleSurf', 'gmSimpleSurf', 'eegPositions')  # TODO: add more keys here once implemented
            for w in allKeys:
                self.clearCache(which=w)
            return

        if which in ('skinSurf', 'csfSurf', 'gmSurf', 'gmSimpleSurf', 'skinSimpleSurf', 'eegPositions'):
            if getattr(self, '_' + which) is None:
                return
            setattr(self, '_' + which, None)
        else:
            raise NotImplementedError()

        self.sigDataChanged.emit(which)

    def _onFilepathChanged(self):
        with self.sigDataChanged.blocked():
            self.clearCache('all')
        self.sigDataChanged.emit(None)

    @property
    def filepath(self):
        return self._filepath

    @filepath.setter
    def filepath(self, newPath: str):
        if self._filepath == newPath:
            return
        self.validateFilepath(newPath)
        self._filepath = newPath
        self.sigFilepathChanged.emit()
        # TODO: here or with slots connected to sigDataChanged, make sure any cached MRI data or metadata is cleared/reloaded

    @property
    def isSet(self):
        return self._filepath is not None

    @property
    def surfKeys(self):
        # TODO: set this dynamically instead of hardcoded
        # TODO: add others
        allSurfKeys = ('skinSurf', 'csfSurf', 'gmSurf')
        return allSurfKeys

    @property
    def gmSurf(self):
        if self.isSet and self._gmSurf is None:
            self.loadCache(which='gmSurf')
        return self._gmSurf

    @property
    def gmSimpleSurf(self):
        """
        Simplified version of gmSurf mesh, for faster plotting.
        """
        if self.isSet and self._gmSimpleSurf is None:
            self.loadCache(which='gmSimpleSurf')
        return self._gmSimpleSurf

    @property
    def csfSurf(self):
        if self.isSet and self._csfSurf is None:
            self.loadCache(which='csfSurf')
        return self._csfSurf

    @property
    def skinSurf(self):
        if self.isSet and self._skinSurf is None:
            self.loadCache(which='skinSurf')
        return self._skinSurf

    @property
    def skinSimpleSurf(self):
        """
        Simplified version of skinSurf mesh, for faster plotting.
        """
        if self.isSet and self._skinSimpleSurf is None:
            self.loadCache(which='skinSimpleSurf')
        return self._skinSimpleSurf

    @property
    def eegPositions(self):
        if self.isSet and self._eegPositions is None:
            self.loadCache(which='eegPositions')
        return self._eegPositions

    def asDict(self, filepathRelTo: str) -> tp.Dict[str, tp.Any]:
        d = attrsAsDict(self)
        if 'filepath' in  d:
            d['filepath'] = os.path.relpath(d['filepath'], filepathRelTo)

        return d

    @classmethod
    def fromDict(cls, d: tp.Dict[str, tp.Any], filepathRelTo: str) -> HeadModel:
        # TODO: validate against schema
        if d.get('filepath', None) is not None:
            d['filepath'] = os.path.join(filepathRelTo, d['filepath'])
            cls.validateFilepath(d['filepath'])
        return cls(**d)

    @classmethod
    def validateFilepath(cls, filepath: tp.Optional[str], strict: bool = False) -> None:
        if filepath is None:
            return
        if strict:
            # make sure .msh actually exists
            assert filepath.endswith('.msh')
            assert os.path.exists(filepath), 'File not found at {}'.format(filepath)
            # TODO: also verify that expected related files (e.g. m2m_* folder) are next to the referenced .msh filepath
        else:
            # just make sure parent directory (typically SimNIBS results dir) actually exists
            parentDir = os.path.dirname(filepath)
            assert os.path.exists(parentDir), 'Parent directory not found at {}'.format(parentDir)



from __future__ import annotations

import asyncio

import attrs
import logging
import numpy as np
import pyvista as pv
import typing as tp
from typing import ClassVar

from . import PlotViewLayer
from NaviNIBS.Navigator.Model.Samples import Sample, Samples
from NaviNIBS.Navigator.Model.Targets import Target, Targets
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.pyvista import DefaultBackgroundPlotter, RemotePlotterProxy
from NaviNIBS.util.pyvista import Actor, setActorUserTransform, concatenateLineSegments


logger = logging.getLogger(__name__)


Transform = np.ndarray


@attrs.define
class VisualizedOrientation:
    """
    Note: this doesn't connect to any change signals from underlying sample, instead assuming that caller will
    re-instantiate for any changes as needed
    """
    _orientation: tp.Union[Sample, Target]
    _plotter: DefaultBackgroundPlotter = attrs.field(repr=False)
    _colorDepthIndicator: str | tuple[float, str]  # string color, or tuple of (float scalar, string colorbar label)
    _colorHandleIndicator: str | tuple[float, str]  # string color, or tuple of (float scalar, string colorbar label)
    _opacity: float
    _lineWidth: float
    _style: str
    _actorKeyPrefix: str

    _actors: tp.Dict[str, Actor] = attrs.field(init=False, factory=dict, repr=False)

    def __attrs_post_init__(self):
        match self._style:
            case 'line':
                raise NotImplementedError  # TODO
            case 'lines':
                # depth axis line plus small handle orientation indicator
                zOffset = -20
                handleLength = 2
                depthLine = pv.lines_from_points(np.asarray([[0, 0, 0], [0, 0, zOffset]]))
                handleLine = pv.lines_from_points(np.asarray([[0, 0, 0], [0, -handleLength, 0]]))

                actorKey = self._actorKeyPrefix + 'depthLine'

                if isinstance(self._colorDepthIndicator, tuple):
                    depthLine[self._colorDepthIndicator[1]] = np.full((depthLine.n_points,), self._colorDepthIndicator[0])
                    scalars = self._colorDepthIndicator[1]
                    scalar_bar_args = {'title': self._colorDepthIndicator[1]}
                    color='k'
                else:
                    scalars = None
                    scalar_bar_args = None
                    color=self._colorDepthIndicator

                self._actors[actorKey] = self._plotter.addLineSegments(
                                                         depthLine,
                                                         name=actorKey,
                                                         color=color,
                                                         scalars=scalars,
                                                         scalar_bar_args=scalar_bar_args,
                                                         width=self._lineWidth,
                                                         opacity=self._opacity,
                                                         userTransform=self._orientation.coilToMRITransf,
                )


                actorKey = self._actorKeyPrefix + 'handleLine'

                if isinstance(self._colorHandleIndicator, tuple):
                    handleLine[self._colorHandleIndicator[1]] = np.full((handleLine.n_points,), self._colorHandleIndicator[0])
                    scalars = self._colorHandleIndicator[1]
                    scalar_bar_args = {'title': self._colorHandleIndicator[1]}
                    color='k'
                else:
                    scalars = None
                    scalar_bar_args = None
                    color=self._colorHandleIndicator

                self._actors[actorKey] = self._plotter.addLineSegments(
                                                         handleLine,
                                                         name=actorKey,
                                                         color=color,
                                                         scalars=scalars,
                                                         scalar_bar_args=scalar_bar_args,
                                                         width=self._lineWidth,
                                                         opacity=self._opacity,
                                                         userTransform=self._orientation.coilToMRITransf,
                )

            case _:
                raise NotImplementedError(f'Unexpected style: {self._style}')

        with self._plotter.allowNonblockingCalls():
            for actor in self._actors.values():
                setActorUserTransform(actor, self._orientation.coilToMRITransf)

    @property
    def actors(self):
        return self._actors


@attrs.define
class OrientationsLayer(PlotViewLayer):
    _type: ClassVar[str] = 'Orientations'
    _colorDepthIndicator: str = '#ba55d3'
    _colorHandleIndicator: str = '#9370db'
    _colorDepthIndicatorSelected: str = '#8b008b'
    _colorHandleIndicatorSelected: str = '#4b0082'
    _opacity: float = 0.5
    _lineWidth: float = 3.
    _style: str = 'lines'

    _visualizedOrientations: tp.Dict[str, VisualizedOrientation] = attrs.field(init=False, factory=dict)
    _loopTask: asyncio.Task = attrs.field(init=False)
    _hasPendingOrientations: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)
    _pendingOrientationKeys: set[str] = attrs.field(init=False, factory=set)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._loopTask = asyncio.create_task(asyncTryAndLogExceptionOnError(self._loop_drawPendingOrientations))

    def _createVisualizedOrientationForSample(self, key: str) -> VisualizedOrientation:
        isSelected = self.orientations[key].isSelected
        return VisualizedOrientation(
            orientation=self.orientations[key],
            plotter=self._plotter,
            colorHandleIndicator=self._colorHandleIndicatorSelected if isSelected else self._colorHandleIndicator,
            colorDepthIndicator=self._colorDepthIndicatorSelected if isSelected else self._colorDepthIndicator,
            opacity=self._opacity,
            lineWidth=self._lineWidth,
            style=self._style,
            actorKeyPrefix=self._getActorKey(key)
        )

    async def _loop_drawPendingOrientations(self):
        """
        Handle drawing of orientations in async loop here, with a yield to make sure we give
        GUI time to handle other tasks during drawing of many orientations (e.g. drawing hundreds of samples)
        """
        while True:
            await asyncio.sleep(0)
            await self._hasPendingOrientations.wait()
            if len(self._pendingOrientationKeys) == 0:
                self._hasPendingOrientations.clear()
                continue

            key = self._pendingOrientationKeys.pop()

            if self._orientationIsVisible(key):
                logger.debug(f'Instantiating visualized orientation for {key}')
                self._visualizedOrientations[key] = self._createVisualizedOrientationForSample(key)
                for actorKey, actor in self._visualizedOrientations[key].actors.items():
                    self._actors[actorKey] = actor
            else:
                logger.debug(f'Skipping drawing of {key} since it should not be visible')

    def _orientationIsVisible(self, key: str) -> bool:
        """
        Whether sample or target should be visible (rendered). Could be because it is marked as always visible, or because it is currently selected.
        """
        raise NotImplementedError  # should be implemented by subclass

    @property
    def orientations(self) -> tp.Union[Samples, Targets]:
        raise NotImplementedError  # should be implemented by subclass

    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str, ...]] = None, changedOrientationKeys: tp.Optional[tp.List[str]] = None):
        super()._redraw(which=which)

        if not isinstance(which, str):
            # assume parent call above triggered appropriate redraws
            return

        if which == 'all':
            which = ['orientations']
            self._redraw(which=which, changedOrientationKeys=changedOrientationKeys)

        elif which == 'orientations':
            if changedOrientationKeys is None:
                changedOrientationKeys = list(set(self._visualizedOrientations.keys()) | set(self.orientations.keys()))

            for key in changedOrientationKeys:
                if key in self._visualizedOrientations:
                    with self._plotter.allowNonblockingCalls():
                        for actorKey in self._visualizedOrientations.pop(key).actors:
                            self._plotter.remove_actor(self._actors.pop(actorKey))

                self._pendingOrientationKeys.add(key)
                self._hasPendingOrientations.set()  # tell orientation drawing loop to check for new key(s) to draw


@attrs.define
class SampleOrientationsLayer(OrientationsLayer):
    _type: ClassVar[str] = 'SampleOrientations'

    _colorDepthIndicator: str = '#ba55d3'
    _colorHandleIndicator: str = '#9370db'
    _colorDepthIndicatorSelected: str = '#8b008b'
    _colorHandleIndicatorSelected: str = '#4b0082'

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._coordinator.session.samples.sigItemsChanged.connect(self._onSamplesChanged)

    @OrientationsLayer.orientations.getter
    def orientations(self) -> Samples:
        return self._coordinator.session.samples

    def _orientationIsVisible(self, key: str) -> bool:
        return key in self._coordinator.session.samples and \
                self._coordinator.session.samples[key].coilToMRITransf is not None and \
               (self._coordinator.session.samples[key].isVisible or
                self._coordinator.session.samples[key].isSelected)

    def _onSamplesChanged(self, changedKeys: tp.List[str], changedAttrs: tp.Optional[tp.List[str]]):
        self._queueRedraw(which='orientations', changedOrientationKeys=changedKeys)


@attrs.define
class TargetOrientationsLayer(OrientationsLayer):
    _type: ClassVar[str] = 'TargetOrientations'

    _colorDepthIndicator: str = '#2a25e3'
    _colorHandleIndicator: str = '#2320eb'
    _colorDepthIndicatorSelected: str = '#2b20ab'
    _colorHandleIndicatorSelected: str = '#2b20a2'
    _lineWidth: float = 4.5

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._coordinator.session.targets.sigItemsChanged.connect(self._onTargetsChanged)

    @OrientationsLayer.orientations.getter
    def orientations(self) -> Targets:
        return self._coordinator.session.targets

    def _orientationIsVisible(self, key: str) -> bool:
        return key in self._coordinator.session.targets and \
               self._coordinator.session.targets[key].coilToMRITransf is not None and \
               (self._coordinator.session.targets[key].isVisible or
                self._coordinator.session.targets[key].isSelected)

    def _onTargetsChanged(self, changedKeys: tp.List[str], changedAttrs: tp.Optional[tp.List[str]]):
        self._queueRedraw(which='orientations', changedOrientationKeys=changedKeys)



import attrs
import logging
from math import ceil
import qtawesome as qta
from qtpy import QtWidgets, QtCore, QtGui
import typing as tp

from NaviNIBS.Devices.ToolPositionsClient import ToolPositionsClient
from NaviNIBS.Navigator.Model.Session import Session, Tool
from NaviNIBS.util.GUI.QFlowLayout import QFlowLayout
from NaviNIBS.util.GUI.IconWidget import IconWidget

logger = logging.getLogger(__name__)


ToolStatusWidget = IconWidget


@attrs.define
class TrackingStatusWidget:
    _session: tp.Optional[Session] = attrs.field(default=None, repr=False)
    _positionsClient: ToolPositionsClient = attrs.field(init=False)
    _wdgt: QtWidgets.QWidget = attrs.field(default=None)
    _hideInactiveTools: bool = True
    _hideToolTypes: tp.List[tp.Type[Tool]] = attrs.field(factory=list)
    _numColumns: int | None = 2  # set to None to use flow layout with variable columns
    _columnContainers: tp.List[QtWidgets.QWidget] = attrs.field(init=False, factory=list)
    _flowLayout: QFlowLayout | None = attrs.field(init=False, default=None)

    _toolWdgts: tp.Dict[str, ToolStatusWidget] = attrs.field(init=False, factory=dict)

    _prevHadTransf: dict[str, bool] = attrs.field(init=False, factory=dict)

    def __attrs_post_init__(self):
        if self._wdgt is None:
            self._wdgt = QtWidgets.QGroupBox('Tools tracking status')

        if self._numColumns is None:
            self._flowLayout = QFlowLayout(layoutMode=QFlowLayout.LayoutMode.equalMinimum)
            # self._flowLayout.sigLayoutChanged.connect(lambda: self._wdgt.adjustSize())
            self._flowLayout.sigLayoutChanged.connect(lambda: self._wdgt.update())
            self._wdgt.setLayout(self._flowLayout)
            self._wdgt.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)

        else:
            self._wdgt.setLayout(QtWidgets.QHBoxLayout())
            self._wdgt.layout().setContentsMargins(0, 0, 0, 0)
            self._wdgt.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
            for iCol in range(self._numColumns):
                wdgt = QtWidgets.QWidget()
                wdgt.setLayout(QtWidgets.QFormLayout())
                self._columnContainers.append(wdgt)
                self._wdgt.layout().addWidget(wdgt)
            self._wdgt.layout().addStretch()

        if self._session is not None:
            self._initialize()

    @property
    def wdgt(self):
        return self._wdgt

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, newSession: tp.Optional[Session]):
        if self._session is newSession:
            return
        self._session = newSession
        self._initialize()

    def _initialize(self):
        if self._session is not None:
            self._positionsClient = ToolPositionsClient()
            self._positionsClient.sigLatestPositionsChanged.connect(self._onLatestPositionsChanged)
            # TODO: connect to any necessary signals from session
            self._session.tools.sigItemsChanged.connect(lambda *args: self._initializeToolWidgets())
            self._initializeToolWidgets()
        else:
            raise NotImplementedError

    def _initializeToolWidgets(self):
        if len(self._toolWdgts) > 0:
            # clean up from previous initialization
            if self._numColumns is None:
                self._flowLayout.clear()
            else:
                for column in self._columnContainers:
                    while (child := column.layout().takeAt(0)) is not None:
                        assert isinstance(child, QtWidgets.QLayoutItem)
                        child.widget().deleteLater()
            self._toolWdgts = {}
            self._prevHadTransf = {}

        toolsToShow = []
        for toolKey, tool in self.session.tools.items():
            if (self._hideInactiveTools and not tool.isActive) \
                    or not tool.doShowTrackingState \
                    or any(isinstance(tool, ToolCls) for ToolCls in self._hideToolTypes):
                continue
            toolsToShow.append(tool)

        for iTool, tool in enumerate(toolsToShow):
            wdgt = IconWidget(icon=qta.icon('mdi6.circle-outline', color='gray'))
            if self._numColumns is None:
                cWdgt = QtWidgets.QWidget()
                cWdgt.setLayout(QtWidgets.QHBoxLayout())
                cWdgt.layout().setContentsMargins(0, 0, 0, 0)
                lWdgt = QtWidgets.QLabel(tool.label)
                lWdgt.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignVCenter)
                cWdgt.layout().addStretch()
                cWdgt.layout().addSpacing(5)
                cWdgt.layout().addWidget(lWdgt)
                cWdgt.layout().addWidget(wdgt)
                cWdgt.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
                self._flowLayout.addWidget(cWdgt)
            else:
                maxNumToolsPerCol = ceil(len(toolsToShow) / self._numColumns)
                iCol = iTool // maxNumToolsPerCol
                self._columnContainers[iCol].layout().addRow(tool.label, wdgt)
            self._toolWdgts[tool.key] = wdgt

        self._onLatestPositionsChanged()

    def _onLatestPositionsChanged(self):
        for toolKey, tool in self.session.tools.items():
            if (self._hideInactiveTools and not tool.isActive) \
                    or not tool.doShowTrackingState \
                    or any(isinstance(tool, ToolCls) for ToolCls in self._hideToolTypes):
                assert toolKey not in self._toolWdgts
                continue

            hasTransf = self._positionsClient.getLatestTransf(tool.trackerKey, None) is not None

            if self._prevHadTransf.get(toolKey, None) in (None, not hasTransf):
                # only update icon if status changed
                wdgt = self._toolWdgts[toolKey]
                if hasTransf:
                    wdgt.icon = qta.icon('mdi6.circle', color='blue')
                else:
                    wdgt.icon = qta.icon('mdi6.help-circle', color='red')
                self._prevHadTransf[toolKey] = hasTransf





from __future__ import annotations
import attrs
import json
import logging
import typing as tp
from typing import TYPE_CHECKING

from NaviNIBS.Navigator.Model.GenericCollection import GenericCollection, GenericCollectionDictItem

if TYPE_CHECKING:
    from qtpy import QtWidgets, QtGui, QtCore

logger = logging.getLogger(__name__)


@attrs.define
class DockWidgetLayout(GenericCollectionDictItem[str]):
    _affinities: list[str]
    _state: dict[str, tp.Any] | None = None
    _winSize: tuple[int, int] | None = None
    """
    Mainly used for root window size, using for other windows may conflict with saved dock states.
    """

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @property
    def affinities(self):
        return self._affinities

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state: dict[str, tp.Any]):
        if self._state == state:
            return
        self.sigItemAboutToChange.emit(self.key, ['state'])
        self._state = state
        self.sigItemChanged.emit(self.key, ['state'])

    @property
    def winSize(self):
        return self._winSize

    @winSize.setter
    def winSize(self, winSize: tuple[int, int] | None):
        if self._winSize == winSize:
            return
        self.sigItemAboutToChange.emit(self.key, ['winSize'])
        self._winSize = winSize
        self.sigItemChanged.emit(self.key, ['winSize'])


@attrs.define
class DockWidgetLayouts(GenericCollection[str, DockWidgetLayout]):
    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @classmethod
    def fromList(cls, itemList: list[dict[str, tp.Any]]) -> DockWidgetLayouts:
        items = {}
        for itemDict in itemList:
            item = DockWidgetLayout.fromDict(itemDict)
            items[item.key] = item

        return cls(items=items)



from __future__ import annotations

import asyncio

import attrs
from datetime import datetime, timedelta
import inspect
import json
import logging
import numpy as np
import os
import pathlib
import pyvista as pv
import pyvistaqt as pvqt
import qtawesome as qta
from qtpy import QtWidgets, QtGui, QtCore
import re
import shutil
import typing as tp

from . import MainViewPanel
from NaviNIBS.Devices.ToolPositionsClient import ToolPositionsClient
from NaviNIBS.Navigator.GUI.Widgets.MRIViews import MRISliceView
from NaviNIBS.Navigator.GUI.Widgets.SurfViews import Surf3DView
from NaviNIBS.Navigator.GUI.Widgets.TrackingStatusWidget import TrackingStatusWidget
from NaviNIBS.Navigator.GUI.Widgets.CollectionTableWidget import HeadPointsTableWidget, RegistrationFiducialsTableWidget
from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.Navigator.Model.SubjectRegistration import Fiducial, HeadPoints
from NaviNIBS.Navigator.Model.Tools import CoilTool, CalibrationPlate
from NaviNIBS.util.pyvista import Actor, setActorUserTransform, RemotePlotterProxy
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.Transforms import applyTransform, invertTransform, transformToString, stringToTransform, estimateAligningTransform, concatenateTransforms
from NaviNIBS.util import makeStrUnique
from NaviNIBS.util import exceptionToStr
from NaviNIBS.util.GUI.ErrorDialog import asyncTryAndRaiseDialogOnError
from NaviNIBS.util.GUI.Icons import getIcon
from NaviNIBS.util.GUI.QFileSelectWidget import QFileSelectWidget
from NaviNIBS.util.GUI.QLineEdit import QLineEditWithValidationFeedback
from NaviNIBS.util.GUI.QTableWidgetDragRows import QTableWidgetDragRows
from NaviNIBS.util.numpy import array_equalish
from NaviNIBS.util.pyvista import DefaultBackgroundPlotter
from NaviNIBS.util.pyvista.dataset import find_closest_point

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@attrs.define
class _PointerDistanceReadout:
    _label: str
    _parentLayout: QtWidgets.QFormLayout
    _units: str = ' mm'
    _value: float = np.nan
    _doHideWhenNaN: bool = False

    _wdgt: tp.Optional[QtWidgets.QLabel] = attrs.field(init=False, default=None)

    def __attrs_post_init__(self):
        self.update()

    def update(self):
        if self._doHideWhenNaN and np.isnan(self._value):
            if self._wdgt is not None:
                labelWdgt = self._parentLayout.labelForField(self._wdgt)
                assert isinstance(labelWdgt, QtWidgets.QLabel)
                labelWdgt.setText('')
                self._wdgt.setText('')
            return

        text = f"{self._value:.1f}{self._units}"
        if self._wdgt is None:
            self._wdgt = QtWidgets.QLabel(text)
            self._parentLayout.addRow(self._label, self._wdgt)
        else:
            self._wdgt.setText(text)

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, newLabel: str):
        self._label = newLabel
        if self._wdgt is not None:
            labelWdgt = self._parentLayout.labelForField(self._wdgt)
            assert isinstance(labelWdgt, QtWidgets.QLabel)
            labelWdgt.setText(self._label)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, newVal: float):
        self._value = newVal
        self.update()


@attrs.define
class _PointerDistanceReadouts:
    _session: Session = attrs.field(repr=False)
    _positionsClient: ToolPositionsClient
    _title: str = 'Current pointer position'

    _wdgt: QtWidgets.QGroupBox = attrs.field(init=False)
    _layout: QtWidgets.QFormLayout = attrs.field(init=False)

    _distToSkinReadout: _PointerDistanceReadout = attrs.field(init=False)
    _distToPlannedFidReadout: _PointerDistanceReadout = attrs.field(init=False)
    _distToSampledFidReadout: _PointerDistanceReadout = attrs.field(init=False)

    def __attrs_post_init__(self):
        self._wdgt = QtWidgets.QGroupBox(self._title)
        self._layout = QtWidgets.QFormLayout()
        self._wdgt.setLayout(self._layout)

        self._distToSkinReadout = _PointerDistanceReadout(
            label='Dist to skin',
            parentLayout=self._layout,
            doHideWhenNaN=False
        )

        self._distToPlannedFidReadout = _PointerDistanceReadout(
            label='Dist to planned fid',
            parentLayout=self._layout,
            doHideWhenNaN=True
        )

        self._distToSampledFidReadout = _PointerDistanceReadout(
            label='Dist to sampled fid',
            parentLayout=self._layout,
            doHideWhenNaN=True
        )

        self._positionsClient.sigLatestPositionsChanged.connect(self._onLatestPositionsChanged)

        # note: could subscribe to fiducial location and head model mesh change signal here, but instead
        # assume that pointer position changes often enough to catch up on any such changes anyway

    @property
    def wdgt(self):
        return self._wdgt

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, newSes: Session):
        self._session = newSes

    def _onLatestPositionsChanged(self):
        self._update()

    def _update(self):
        pointerToCameraTransf = self._positionsClient.getLatestTransf(self.session.tools.pointer.trackerKey, None)
        subjectTrackerToCameraTransf = self._positionsClient.getLatestTransf(self.session.tools.subjectTracker.trackerKey, None)

        if pointerToCameraTransf is None or subjectTrackerToCameraTransf is None:
            # TODO: report NaNs for all distances
            return

        pointerCoord_relToSubTracker = applyTransform([self.session.tools.pointer.toolToTrackerTransf,
                                                       pointerToCameraTransf,
                                                       invertTransform(subjectTrackerToCameraTransf)
                                                       ], np.zeros((3,)), doCheck=False)

        subjectTrackerToMRITransf = self.session.subjectRegistration.trackerToMRITransf
        if subjectTrackerToMRITransf is None:
            # TODO: report NaNS for all distances
            return

        pointerCoord_MRISpace = applyTransform(subjectTrackerToMRITransf, pointerCoord_relToSubTracker, doCheck=False)

        # find distance to skin
        closestPtIndex = find_closest_point(self.session.headModel.skinSurf, pointerCoord_MRISpace)
        closestPt = self.session.headModel.skinSurf.points[closestPtIndex, :]
        dist = np.linalg.norm(closestPt - pointerCoord_MRISpace)
        self._distToSkinReadout.value = dist

        # determine which fiducial to compare to
        plannedFiducialCoords = self.session.subjectRegistration.fiducials.plannedFiducials
        sampledFiducialCoords = self.session.subjectRegistration.fiducials.sampledFiducials

        # look at both planned and sampled fiducials to choose single closest, then use that for both planned and sampled

        whichFidClosest: tp.Optional[str] = None
        closestDist = np.inf

        for whichType, coords in (('planned', plannedFiducialCoords), ('sampled', sampledFiducialCoords)):
            for whichFid, coord in coords.items():
                if coord is None:  # e.g. fiducial not yet sampled
                    continue
                if whichType == 'sampled':
                    # must do extra coordinate conversion from head tracker space to MRI space
                    coord = applyTransform(subjectTrackerToMRITransf, coord, doCheck=False)
                dist = np.linalg.norm(coord - pointerCoord_MRISpace)
                if dist < closestDist:
                    closestDist = dist
                    whichFidClosest = whichFid

        if False:
            # TODO: debug, delete
            whichFidClosest = list(sampledFiducialCoords.keys())[-1]

        if closestDist > 50:
            # don't show any fiducial distance feedback if not close to any fiducials
            whichFidClosest = None

        if whichFidClosest is None:
            # hide fiducial distance readouts
            self._distToPlannedFidReadout.value = np.nan
            self._distToSampledFidReadout.value = np.nan

        else:
            coord = plannedFiducialCoords[whichFidClosest]
            if coord is None:
                self._distToPlannedFidReadout.value = np.nan
            else:
                self._distToPlannedFidReadout.value = np.linalg.norm(coord - pointerCoord_MRISpace)
                self._distToPlannedFidReadout.label = f'Dist to planned {whichFidClosest}'

            coord = sampledFiducialCoords.get(whichFidClosest, None)
            if coord is None:
                self._distToSampledFidReadout.value = np.nan
            else:
                # must do extra coordinate conversion from head tracker space to MRI space
                coord = applyTransform(subjectTrackerToMRITransf, coord, doCheck=False)

                self._distToSampledFidReadout.value = np.linalg.norm(coord - pointerCoord_MRISpace)
                self._distToSampledFidReadout.label = f'Dist to sampled {whichFidClosest}'


@attrs.define
class SubjectRegistrationPanel(MainViewPanel):
    """
    SubjectRegistrationPanel

    Note: need to carefully manage sequencing / interaction between fiducials and head points.
    Example scenario:
    1. Plan fiducials
    2. Sample planned fiducials
    3. Align fiducials
    4. Sample head points
    5. Refine alignment with head points
    6. Convert sampled fiducials to planned
    7. Create another fiducial from pointer position (e.g. head ref)
    Some time later, tracker moves and we must reregister:
    8. Re-sample (sampled-converted-to-planned) fiducials and head ref
    9. Align fiducials
    10. Don't resample head points, but expect them to update their alignment
    This last point is what requires careful attention in sequencing.
    Head points are defined in tracker space. If the tracker moves, the old points
    are no longer valid (but we may want to update them for the new alignment for display purposes).

    Another example scenario, not using sample->planned fiducial conversion:
    1. Plan fiducials
    2. Sample planned fiducials
    3. Align fiducials
    4. Sample head points
    5. Refine alignment with head points
    Some time later, tracker moves and we must reregister:
    8. Re-sample planned fiducials
    9. Align fiducials
    10. Don't resample head points, but expect them to update their alignment(?)

    But also must allow this scenario:
    1. Plan fiducials
    2. Sample planned fiducials
    3. Align fiducials
    4. Sample head points
    5. Refine alignment with head points
    6. Align to fiducials again to "undo" head point refinement
    7. Change a refinement setting (e.g. weight) and refine again, without needing to resample head points

    Less common scenario that would also be nice to support:
    1. Plan fiducials
    2. Sample planned fiducials
    3. Align fiducials
    4. Sample head points
    5. Refine alignment with head points
    6. Edit a *planned* fiducial to make it align better with reality.
    7. Align fiducials.
    8. Refine alignment with head points, without needing to resample.

    So, if re-aligning to fiducials but *sampled* fiducials haven't actually changed, we should assume that tracker position hasn't changed, and so head points don't need to be adjusted.

    But if fiducials have changed, and then we re-align to those fiducials, then we should assume
        that the tracker position did change. If we do nothing, head points will then be
        incorrect/nonnsensical. We must either convert head points into the new tracker space,
        or delete them.

    To convert head points into new tracker space, can we assume that previous tracker to MRI
    transform was "correct", and go through that?

    TODO: maybe use fiducial history or add a "last edited" timestamp to fiducial fields and
    head points to not be responsible for tracking this sequencing here.
    """
    _key: str = 'Register'
    _icon: QtGui.QIcon = attrs.field(init=False, factory=lambda: getIcon('mdi6.head-snowflake'))
    _surfKey: str = 'skinSurf'

    _trackingStatusWdgt: TrackingStatusWidget = attrs.field(init=False)
    _fidTblWdgt: RegistrationFiducialsTableWidget = attrs.field(init=False)
    _headPtsTblWdgt: HeadPointsTableWidget = attrs.field(init=False)
    _sampleFiducialBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _sampleFiducialMultipleBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _clearFiducialBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _newFidFromPointerBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _newFidFromSampleBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _alignToFiducialsBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _sampleHeadPtsBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _clearHeadPtsBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _refineWeightsField: QLineEditWithValidationFeedback = attrs.field(init=False)
    _refineWithHeadpointsBtn: QtWidgets.QPushButton = attrs.field(init=False)
    _plotter: DefaultBackgroundPlotter = attrs.field(init=False)
    _actors: tp.Dict[str, tp.Optional[Actor]] = attrs.field(init=False, factory=dict)
    _pointerDistanceReadouts: _PointerDistanceReadouts = attrs.field(init=False)

    _positionsClient: tp.Optional[ToolPositionsClient] = attrs.field(init=False, default=None)

    finishedAsyncInitializationEvent: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def canBeEnabled(self) -> tuple[bool, str | None]:
        if self.session is None:
            return False, 'No session set'
        if not self.session.MRI.isSet:
            return False, 'No MRI set'
        if not self.session.headModel.isSet:
            return False, 'No head model set'
        if self.session.tools.subjectTracker is None:
            return False, 'No active subject tracker configured'
        if self.session.tools.pointer is None:
            return False, 'No active pointer tool configured'
        if not self.session.subjectRegistration.hasMinimumPlannedFiducials:
            return False, 'Missing planned fiducials'
        return True, None

    def _finishInitialization(self):
        super()._finishInitialization()

        self._positionsClient = ToolPositionsClient()
        self._positionsClient.sigLatestPositionsChanged.connect(lambda: self._redraw(which=['pointerPosition', 'sampleBtns']))

        self._wdgt.setLayout(QtWidgets.QHBoxLayout())
        self._wdgt.layout().setContentsMargins(0, 0, 0, 0)

        sidebar = QtWidgets.QWidget()
        sidebar.setLayout(QtWidgets.QVBoxLayout())
        sidebar.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Expanding)
        self._wdgt.layout().addWidget(sidebar)

        self._trackingStatusWdgt = TrackingStatusWidget(session=self.session,
                                                        hideToolTypes=[CoilTool, CalibrationPlate])
        sidebar.layout().addWidget(self._trackingStatusWdgt.wdgt)

        fiducialsBox = QtWidgets.QGroupBox('Fiducials')
        fiducialsBox.setLayout(QtWidgets.QVBoxLayout())
        sidebar.layout().addWidget(fiducialsBox)

        btnContainer = QtWidgets.QWidget()
        btnContainer.setLayout(QtWidgets.QGridLayout())
        btnContainer.layout().setContentsMargins(0, 0, 0, 0)
        fiducialsBox.layout().addWidget(btnContainer)

        btn = QtWidgets.QPushButton('Sample fiducial (single)')
        btn.clicked.connect(self._onSampleFidBtnClicked)
        self._sampleFiducialBtn = btn
        btnContainer.layout().addWidget(btn, 0, 0, )

        btn = QtWidgets.QPushButton('Sample fiducial (multiple)')
        btn.clicked.connect(self._onSampleFidMultipleBtnClicked)
        self._sampleFiducialMultipleBtn = btn
        btnContainer.layout().addWidget(btn, 1, 0, )

        btn = QtWidgets.QPushButton('Clear fiducial')
        btn.clicked.connect(self._onClearFidBtnClicked)
        btnContainer.layout().addWidget(btn, 0, 1)
        # TODO: change this to "clear fiducials" when multiple selected
        self._clearFiducialBtn = btn

        btn = QtWidgets.QPushButton('New planned fiducial from pointer')
        btn.clicked.connect(self._onNewFidFromPointerClicked)
        btnContainer.layout().addWidget(btn, 2, 0, 1, 2)
        btn.setEnabled(False)
        self._newFidFromPointerBtn = btn

        btn = QtWidgets.QPushButton('New planned fiducial from sample')
        btn.clicked.connect(self._onNewFidFromSampleClicked)
        btnContainer.layout().addWidget(btn, 3, 0, 1, 2)
        self._newFidFromSampleBtn = btn

        # TODO: add prev/next fiducial buttons for mapping to foot pedal actions
        # (i.e. without requiring click in fidTbl to select different fiducial)

        self._fidTblWdgt = RegistrationFiducialsTableWidget()
        self._fidTblWdgt.sigSelectionChanged.connect(self._onSelectedFiducialsChanged)
        self._fidTblWdgt.wdgt.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        fiducialsBox.layout().addWidget(self._fidTblWdgt.wdgt)

        # TODO: separate fiducials and samples into an adjustable splitter

        btnContainer = QtWidgets.QWidget()
        btnContainer.setLayout(QtWidgets.QGridLayout())
        btnContainer.layout().setContentsMargins(0, 0, 0, 0)
        fiducialsBox.layout().addWidget(btnContainer)

        btn = QtWidgets.QPushButton('Align to sampled fiducials')
        btn.clicked.connect(self._onAlignToFidBtnClicked)
        self._alignToFiducialsBtn = btn
        btnContainer.layout().addWidget(btn, 0, 0, 1, 2)

        headPtsBox = QtWidgets.QGroupBox('Head shape points')
        headPtsBox.setLayout(QtWidgets.QVBoxLayout())
        sidebar.layout().addWidget(headPtsBox)

        btnContainer = QtWidgets.QWidget()
        btnContainer.setLayout(QtWidgets.QGridLayout())
        btnContainer.layout().setContentsMargins(0, 0, 0, 0)
        headPtsBox.layout().addWidget(btnContainer)

        btn = QtWidgets.QPushButton('Sample head point')
        btn.clicked.connect(self._onSampleHeadPtsBtnClicked)
        btn = self._sampleHeadPtsBtn = btn
        btnContainer.layout().addWidget(btn, 0, 0)

        btn = QtWidgets.QPushButton('Delete selected point')
        btn.clicked.connect(self._onClearHeadPtsBtnClicked)
        btnContainer.layout().addWidget(btn, 0, 1)
        btn.setEnabled(False)
        self._clearHeadPtsBtn = btn

        self._headPtsTblWdgt = HeadPointsTableWidget(doAdjustSizeToContents=False)
        self._headPtsTblWdgt.sigSelectionChanged.connect(self._onSelectedHeadPointsChanged)
        self._headPtsTblWdgt.wdgt.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)

        headPtsBox.layout().addWidget(self._headPtsTblWdgt.wdgt)

        container = QtWidgets.QWidget()
        container.setLayout(QtWidgets.QFormLayout())
        container.layout().setContentsMargins(0, 0, 0, 0)
        headPtsBox.layout().addWidget(container)
        self._refineWeightsField = QLineEditWithValidationFeedback()
        container.layout().addRow('Refine weights', self._refineWeightsField)
        self._refineWeightsField.setToolTip(
            inspect.cleandoc(
                """
                Optional weights used for headpoint-based registration refinement, in format expected by 
                simpleicp's `rbp_observation_weights` argument (i.e. rot_x, rot_y, rot_z, t_x, t_y, t_z). 
                Can alternatively specify as a single scalar to apply the same weight to all terms, or as 
                two scalars to apply the first weight to all rotation term, the second weights to all
                translation terms.
                
                Note that values will be inverted (weightArg = 1 / weight) so that higher input values 
                correspond to greater weighting of head points, rather than greater weighting of fiducials.
                
                Note that these weights are different in definition and usage than `Fiducial.alignmentWeight`.
                """
            )
        )
        self._refineWeightsField.setPlaceholderText('Optional')
        self._refineWeightsField.editingFinished.connect(self._onHeadPointsWeightsFieldChanged)

        class RefineWeightsValidator(QtGui.QValidator):
            def __init__(self):
                super().__init__()
                self._invalidRegex = re.compile(r'(\[\[)|([a-zA-Z]+)|(\d+ +[\d.]+)|(]])|(,,)|(\.\.)|(] +\[)')

            def validate(self, inputStr: str, pos: int) -> QtGui.QValidator.State:
                if len(inputStr) == 0:
                    return self.State.Acceptable, inputStr, pos
                try:
                    listOrScalarVal = json.loads(inputStr)
                    if not isinstance(listOrScalarVal, (list, float, int)):
                        raise ValueError('Expected list or scalar')
                    arrayVal = np.asarray(listOrScalarVal, dtype=np.float64)  # raises value error if problem
                except ValueError as e:
                    if self._invalidRegex.search(inputStr):
                        return self.State.Invalid, inputStr, pos
                    else:
                        return self.State.Intermediate, inputStr, pos
                else:
                    return self.State.Acceptable, inputStr, pos

        self._refineWeightsField.setValidator(RefineWeightsValidator())

        btnContainer = QtWidgets.QWidget()
        btnContainer.setLayout(QtWidgets.QGridLayout())
        btnContainer.layout().setContentsMargins(0, 0, 0, 0)
        headPtsBox.layout().addWidget(btnContainer)

        btn = QtWidgets.QPushButton('Refine with sampled head points')
        btn.clicked.connect(self._onAlignToHeadPtsBtnClicked)
        btn.setEnabled(False)
        self._refineWithHeadpointsBtn = btn
        btnContainer.layout().addWidget(btn, 0, 0, 1, 2)

        self._pointerDistanceReadouts = _PointerDistanceReadouts(
            session=self.session,
            positionsClient=self._positionsClient,
        )
        sidebar.layout().addWidget(self._pointerDistanceReadouts.wdgt)

        self._plotter = DefaultBackgroundPlotter()
        self._wdgt.layout().addWidget(self._plotter)

        asyncio.create_task(asyncTryAndRaiseDialogOnError(self._finishInitialization_async))

    async def _finishInitialization_async(self):
        if isinstance(self._plotter, RemotePlotterProxy):
            await self._plotter.isReadyEvent.wait()

        self._plotter.enable_depth_peeling(4)

        if self.session is not None:
            self._onPanelInitializedAndSessionSet()

        self.finishedAsyncInitializationEvent.set()

    def _onSessionSet(self):
        super()._onSessionSet()
        # TODO: connect relevant session changed signals to _redraw calls

        if self._hasInitialized:
            self._onPanelInitializedAndSessionSet()

    def _onPanelInitializedAndSessionSet(self):
        self.session.headModel.sigDataChanged.connect(lambda which: self._redraw(which='initSurf'))
        self.session.subjectRegistration.fiducials.sigItemsChanged.connect(self._onFiducialsChanged)
        self.session.subjectRegistration.sampledHeadPoints.sigHeadpointsChanged.connect(self._onHeadpointsChanged)
        self.session.subjectRegistration.sampledHeadPoints.sigAttribsChanged.connect(self._onHeadPointAttribsChanged)
        self.session.subjectRegistration.sigTrackerToMRITransfChanged.connect(lambda: self._redraw(which=[
            'initSampledFids', 'initHeadPts', 'initSubjectTracker', 'initPointer', 'pointerPosition']))
        self._updateHeadPointsWeightsField()

        self._trackingStatusWdgt.session = self.session

        self._fidTblWdgt.session = self.session
        self._fidTblWdgt.currentRow = 0
        self._headPtsTblWdgt.session = self.session

        self._pointerDistanceReadouts.session = self.session

        self._redraw(which='all')

    def _onFiducialsChanged(self, fidKeys: list[str], attribs: tp.Optional[list[str]] = None):
        # TODO: pass fidKeys to redraw below to only redraw changed fiducials
        if attribs is None or 'plannedCoord' in attribs:
            self._redraw(which='initPlannedFids')
        if attribs is None or 'sampledCoord' in attribs or 'sampledCoords' in attribs:
            self._redraw(which=['initSampledFids', 'alignBtn'])

    def _onHeadpointsChanged(self, ptIndices: list[int], attribs: tp.Optional[list[str]] = None):
        self._redraw(which='initHeadPts')  # TODO: pass which indices to redraw to not entirely redraw each time there's a change
        self._onSelectedHeadPointsChanged(self._headPtsTblWdgt.selectedCollectionItemKeys)

    def _onHeadPointAttribsChanged(self, attribKeys: list[str]):
        """
        Called when an attribute of the headpoints instance changes other than the head points themselves.

        For now, this is just the alignmentWeights, but could be something else in the future.
        """
        if 'alignmentWeights' in attribKeys:
            self._updateHeadPointsWeightsField()

    def _updateHeadPointsWeightsField(self):
        if self.session is None:
            self._refineWeightsField.setText('')
        else:
            weights = self.session.subjectRegistration.sampledHeadPoints.alignmentWeights
            if weights is None:
                self._refineWeightsField.setText('')
            else:
                newText = json.dumps(weights.tolist())
                if self._refineWeightsField.text() != newText:
                    self._refineWeightsField.setText(newText)

    def _onHeadPointsWeightsFieldChanged(self):
        try:
            weightsStr = self._refineWeightsField.text()
            if len(weightsStr)==0:
                weights = None
            else:
                weights = json.loads(weightsStr)
                weights = np.atleast_1d(np.asarray(weights)).astype(np.float64)
            self.session.subjectRegistration.sampledHeadPoints.alignmentWeights = weights
        except Exception as e:
            # revert text
            logger.warning(f'Error trying to set weights from str {weightsStr}: {exceptionToStr(e)}')
            self._updateHeadPointsWeightsField()

    def _currentFidTblFidKey(self) -> tp.Optional[str]:
        return self._fidTblWdgt.currentCollectionItemKey

    def _getPointerCoordRelToSubTracker(self) -> tp.Optional[np.ndarray]:
        # TODO: spin-wait update positionsClient.latestPositions to make sure we have the most up to date position

        pointerToCameraTransf = self._positionsClient.getLatestTransf(self.session.tools.pointer.trackerKey, None)
        subjectTrackerToCameraTransf = self._positionsClient.getLatestTransf(self.session.tools.subjectTracker.trackerKey, None)

        if pointerToCameraTransf is None or subjectTrackerToCameraTransf is None:
            logger.warning('Tried to sample, but do not have valid positions. Returning.')
            return None

        #logger.info('Sampled fiducial:\npointer: {}\ntracker: {}'.format(pointerToCameraTransf, subjectTrackerToCameraTransf))

        pointerCoord_relToSubTracker = applyTransform([self.session.tools.pointer.toolToTrackerTransf,
                                                       pointerToCameraTransf,
                                                       invertTransform(subjectTrackerToCameraTransf)
                                                       ], np.zeros((3,)), doCheck=False)

        return pointerCoord_relToSubTracker

    def _onSampleFidBtnClicked(self):
        fidKey = self._currentFidTblFidKey()

        pointerCoord_relToSubTracker = self._getPointerCoordRelToSubTracker()
        if pointerCoord_relToSubTracker is None:
            return

        logger.info(f'Sampled fiducial relative to tracker: {pointerCoord_relToSubTracker}')

        self.session.subjectRegistration.fiducials[fidKey].sampledCoord = pointerCoord_relToSubTracker

        if True:
            currentRow = self._fidTblWdgt.currentRow
            if currentRow == self._fidTblWdgt.rowCount - 1:
                # already at end of table
                # TODO: auto-advance to prompt about aligning
                pass
            else:
                # advance to next fiducial in table
                self._fidTblWdgt.currentRow += 1

    def _onSampleFidMultipleBtnClicked(self):
        fidKey = self._currentFidTblFidKey()

        pointerCoord_relToSubTracker = self._getPointerCoordRelToSubTracker()
        if pointerCoord_relToSubTracker is None:
            return

        logger.info(f'Sampled fiducial relative to tracker: {pointerCoord_relToSubTracker}')

        coords = self.session.subjectRegistration.fiducials[fidKey].sampledCoords
        if coords is None:
            coords = pointerCoord_relToSubTracker[np.newaxis, :]
        else:
            coords = np.vstack((coords, pointerCoord_relToSubTracker[np.newaxis, :]))
        self.session.subjectRegistration.fiducials[fidKey].sampledCoords = coords

        # don't auto-advance to allow additional samples of this fiducial

    def _onClearFidBtnClicked(self):
        fidKeys = self._getSelectedFiducialKeys()
        logger.info(f'Clearing sampled fiducials {fidKeys}')
        for fidKey in fidKeys:
            self.session.subjectRegistration.fiducials[fidKey].sampledCoords = None

    def _onNewFidFromPointerClicked(self):
        sampledCoord = self._getPointerCoordRelToSubTracker()
        if sampledCoord is None:
            logger.warning('Cannot create fiducial from pointer when pointer is not visible')
            return

        if self.session.subjectRegistration.trackerToMRITransf is None:
            logger.warning('Cannot create fiducial from pointer when trackerToMRITransf is unknown')
            return

        plannedCoord = applyTransform(self.session.subjectRegistration.trackerToMRITransf, sampledCoord)

        newFidKey = makeStrUnique('SampledFiducial',
                                  existingStrs=self.session.subjectRegistration.fiducials.keys(),
                                  delimiter='')

        logger.info(f'Creating new fiducial {newFidKey} from current pointer position')

        self.session.subjectRegistration.fiducials.addItem(
            Fiducial(
                key=newFidKey,
                plannedCoord=plannedCoord,
                sampledCoords=sampledCoord
            )
        )

    def _onNewFidFromSampleClicked(self):

        if self.session.subjectRegistration.trackerToMRITransf is None:
            logger.warning('Cannot create fiducial from pointer when trackerToMRITransf is unknown')
            return

        selFidKeys = self._getSelectedFiducialKeys()

        for selFidKey in selFidKeys:
            sampledCoord = self.session.subjectRegistration.fiducials[selFidKey].sampledCoord
            if sampledCoord is None:
                # this fiducial doesn't have a sampled coordinate
                continue

            plannedCoord = applyTransform(self.session.subjectRegistration.trackerToMRITransf, sampledCoord)

            newFidKey = makeStrUnique(f'Sampled{selFidKey}',
                                      existingStrs=self.session.subjectRegistration.fiducials.keys(),
                                      delimiter='')

            logger.info(f'Creating new fiducial {newFidKey} from sampled fiducial {selFidKey} position')

            self.session.subjectRegistration.fiducials.addItem(
                Fiducial(
                    key=newFidKey,
                    plannedCoord=plannedCoord,
                    sampledCoords=sampledCoord
                )
            )

            # clear sampled positions of previously selected fiducials so that they're not automatically used for future alignments
            self.session.subjectRegistration.fiducials[selFidKey].sampledCoords = None


    def _onFidTblCurrentCellChanged(self, currentRow: int, currentCol: int, previousRow: int, previousCol: int):
        if previousRow == currentRow:
            return  # no change in row selection

        self._redraw(which=['sampleBtns'])

        fidKey = self._currentFidTblFidKey()

        if fidKey is None:
            # no selection (perhaps table was temporarily cleared)
            return

        subReg = self.session.subjectRegistration
        if fidKey in subReg.fiducials and subReg.fiducials[fidKey].plannedCoord is not None:
            lookAt = subReg.fiducials[fidKey].plannedCoord
        elif fidKey in subReg.fiducials and subReg.fiducials[fidKey].sampledCoord is not None and subReg.trackerToMRITransf is not None:
            lookAt = applyTransform(subReg.trackerToMRITransf, subReg.fiducials[fidKey].sampledCoord)
        else:
            lookAt = None
        if lookAt is not None:
            self._plotter.camera.focal_point = lookAt
            vec = lookAt - subReg.approxHeadCenter
            self._plotter.camera.position = lookAt + vec*10
            self._plotter.reset_camera()

    def _getSelectedFiducialKeys(self):
        return self._fidTblWdgt.selectedCollectionItemKeys

    def _onSelectedFiducialsChanged(self, selKeys: list[str]):
        if len(selKeys) == 0:
            self._clearFiducialBtn.setEnabled(False)
            self._clearFiducialBtn.setText('Clear fiducial')
            self._newFidFromSampleBtn.setEnabled(False)
            self._newFidFromSampleBtn.setText('New planned fiducial from sample')
        else:
            numSelFiducialsWithSamples = sum(self.session.subjectRegistration.fiducials[key].sampledCoord is not None for key in selKeys)

            if len(selKeys) == 1:
                self._clearFiducialBtn.setEnabled(True)
                self._clearFiducialBtn.setText('Clear fiducial')
            else:
                self._clearFiducialBtn.setEnabled(True)
                self._clearFiducialBtn.setText('Clear fiducials')

            if numSelFiducialsWithSamples == 0:
                self._newFidFromSampleBtn.setEnabled(False)
            elif numSelFiducialsWithSamples == 1:
                self._newFidFromSampleBtn.setEnabled(True)
                self._newFidFromSampleBtn.setText('New planned fiducial from sample')
            else:
                self._newFidFromSampleBtn.setEnabled(True)
                self._newFidFromSampleBtn.setText('New planned fiducials from samples')

    def _onFidTblCellDoubleClicked(self):
        raise NotImplementedError()  # TODO

    def _onAlignToFidBtnClicked(self):

        subReg = self.session.subjectRegistration
        assert subReg.hasMinimumSampledFiducials

        validPlannedFidKeys = {key for key, fid in subReg.fiducials.items() if fid.plannedCoord is not None}
        validSampledFidKeys = {key for key, fid in subReg.fiducials.items() if fid.sampledCoord is not None}
        commonKeys = validPlannedFidKeys & validSampledFidKeys
        assert len(commonKeys) >= 3

        plannedPts_mriSpace = np.vstack([subReg.fiducials[key].plannedCoord for key in commonKeys])
        sampledPts_subSpace = np.vstack([subReg.fiducials[key].sampledCoord for key in commonKeys])

        alignmentWeights = np.asarray([subReg.fiducials[key].alignmentWeight for key in commonKeys])

        if all(alignmentWeights == 1.):
            alignmentWeights = None

        logger.info('Estimating transform aligning sampled fiducials to planned fiducials')
        newTrackerToMRITransf = estimateAligningTransform(
            sampledPts_subSpace,
            plannedPts_mriSpace,
            weights=alignmentWeights)

        if len(subReg.sampledHeadPoints) > 0:
            # TODO: handle updating coordinate system of sampled head points if needed
            # did tracker move since last alignment?
            # if sampled fiducials were updated since trackerToMRITransf last updated,
            # then we should assume that the tracker moved and we need to update head points

            timeOfLastSampledFiducialChange = None
            # (assumes history is in chronological order, with most recent last)

            historyTimes = list(subReg.fiducialsHistory.keys())
            historyDatetimes = [subReg.getDatetimeFromTimestamp(t) for t in historyTimes]

            if len(subReg.fiducialsHistory) > 0:
                assert subReg.fiducialsHistory[historyTimes[-1]] == subReg.fiducials, \
                    'last item in history does not match current state'

            for iHist in range(len(historyTimes) - 2, -1, -1):
                timeDiff = historyDatetimes[iHist + 1] - historyDatetimes[iHist]
                assert timeDiff > timedelta(0), 'fiducial history times are not in chronological order'

                fids_i = subReg.fiducialsHistory[historyTimes[iHist]].sampledFiducials
                fids_j = subReg.fiducialsHistory[historyTimes[iHist + 1]].sampledFiducials

                fidKeys_i = set(fids_i.keys())
                for fidKey_j, fidCoord_j in fids_j.items():
                    if fidKey_j not in fids_i:
                        # new fiducial
                        timeOfLastSampledFiducialChange = historyDatetimes[iHist + 1]
                        break
                    else:
                        fidKeys_i.remove(fidKey_j)
                        if not array_equalish(fidCoord_j, fids_i[fidKey_j]):
                            # fiducial changed
                            timeOfLastSampledFiducialChange = historyDatetimes[iHist + 1]
                            break

                if timeOfLastSampledFiducialChange is None:
                    if len(fidKeys_i) > 0:
                        # previously-sampled fiducial is now removed
                        timeOfLastSampledFiducialChange = historyDatetimes[iHist + 1]
                        break

                if timeOfLastSampledFiducialChange is not None:
                    break

            if len(subReg.trackerToMRITransfHistory) > 0:
                timeOfLastTrackerToMRITransfChange = subReg.getDatetimeFromTimestamp(list(subReg.trackerToMRITransfHistory.keys())[-1])
            else:
                timeOfLastTrackerToMRITransfChange = None

            if timeOfLastSampledFiducialChange is None:
                trackerMoved = False
            elif timeOfLastTrackerToMRITransfChange is None:
                trackerMoved = True
            elif timeOfLastTrackerToMRITransfChange > timeOfLastSampledFiducialChange:
                trackerMoved = False
            else:
                trackerMoved = True

            if trackerMoved:

                headPtsTransf = concatenateTransforms([subReg.trackerToMRITransf, invertTransform(newTrackerToMRITransf)])

                logger.info(f'Applying transform to align previously sampled head points to new tracker position: {headPtsTransf}')

                subReg.sampledHeadPoints.replace(applyTransform(headPtsTransf, np.asarray(subReg.sampledHeadPoints)))

        subReg.trackerToMRITransf = newTrackerToMRITransf

    def _onSampleHeadPtsBtnClicked(self):

        pointerCoord_relToSubTracker = self._getPointerCoordRelToSubTracker()
        if pointerCoord_relToSubTracker is None:
            return

        logger.info(f'Sampled head pt relative to tracker: {pointerCoord_relToSubTracker}')

        self.session.subjectRegistration.sampledHeadPoints.append(pointerCoord_relToSubTracker)

    def _onClearHeadPtsBtnClicked(self):
        self.session.subjectRegistration.sampledHeadPoints.remove(self._headPtsTblWdgt.selectedCollectionItemKeys)

    def _onSelectedHeadPointsChanged(self, selectedIndices: list[int]):
        numSelHeadPts = len(self._headPtsTblWdgt.selectedCollectionItemKeys)
        if numSelHeadPts > 0:
            self._clearHeadPtsBtn.setEnabled(True)
            if numSelHeadPts == 1:
                self._clearHeadPtsBtn.setText('Delete selected point')
            else:
                self._clearHeadPtsBtn.setText('Delete selected points')
        else:
            self._clearHeadPtsBtn.setEnabled(False)

        self._redraw(which='initHeadPts')  # TODO: redraw just previously and currently selected points instead of all

    def _onAlignToHeadPtsBtnClicked(self):
        sampledHeadPts_trackerSpace = np.asarray(self.session.subjectRegistration.sampledHeadPoints)
        sampledHeadPts_MRISpace = applyTransform(self.session.subjectRegistration.trackerToMRITransf, sampledHeadPts_trackerSpace)

        meshHeadPts_MRISpace = self.session.headModel.skinSurf.points

        alignmentWeights = self.session.subjectRegistration.sampledHeadPoints.alignmentWeights
        if alignmentWeights is not None:
            match len(alignmentWeights):
                case 1:
                    alignmentWeights = np.tile(alignmentWeights, (6,))
                case 2:
                    assert alignmentWeights.ndim == 1
                    alignmentWeights = np.tile(alignmentWeights, (3, 1)).T.flatten()
                case 6:
                    pass  # no changes
                case _:
                    raise NotImplementedError('Unexpected size of alignmentWeights')

            # Invert weights so that larger input weights correspond to less trusting of initial values.
            #  Also, convert input zero values to inf values in output.
            icpObservationWeights = np.divide(1, alignmentWeights,
                                              out=np.full_like(alignmentWeights, np.inf),
                                              where=alignmentWeights != 0)

        else:
            icpObservationWeights = None

        # TODO: offload this to a separate thread/process so it doesn't block main GUI
        # (but then also need to implement progress bar, option to cancel, locking of other dependent GUI controls)
        extraTransf = estimateAligningTransform(sampledHeadPts_MRISpace, meshHeadPts_MRISpace,
                                                method='ICP',
                                                weights=icpObservationWeights)

        logger.info(f'Extra transf from refinining head points: {extraTransf}')

        self.session.subjectRegistration.trackerToMRITransf = concatenateTransforms([self.session.subjectRegistration.trackerToMRITransf, extraTransf])

    def _redraw(self, which: tp.Union[str, tp.List[str,...]]):

        logger.debug('redraw {}'.format(which))

        if isinstance(self._plotter, RemotePlotterProxy) and not self._plotter.isReadyEvent.is_set():
            # plotter not ready yet
            return

        if isinstance(which, list):
            for subWhich in which:
                self._redraw(which=subWhich)
            return

        if which == 'all':
            which = ['initSurf', 'initSubjectTracker', 'initPointer', 'initPlannedFids', 'initSampledFids', 'initHeadPts',
                     'sampleBtns', 'alignBtn']
            self._redraw(which=which)
            return

        elif which == 'initSurf':
            actorKey = 'surf'
            self._actors[actorKey] = self._plotter.add_mesh(mesh=getattr(self.session.headModel, self._surfKey),
                                                          color='#d9a5b2',
                                                          opacity=0.8,  # TODO: make GUI-configurable
                                                          name=actorKey)

        elif which == 'initSubjectTracker':

            subjectTracker = self.session.tools.subjectTracker
            doShowSubjectTracker = self.session.subjectRegistration.trackerToMRITransf is not None \
                                and subjectTracker is not None \
                                and subjectTracker.trackerSurf is not None
            actorKey = 'subjectTracker'
            if not doShowSubjectTracker:
                if actorKey in self._actors:
                    self._plotter.remove_actor(self._actors[actorKey])
                    self._actors.pop(actorKey)

            else:
                self._actors[actorKey] = self._plotter.addMesh(mesh=self.session.tools.subjectTracker.trackerSurf,
                                                               color=self.session.tools.subjectTracker.trackerColor,
                                                               defaultMeshColor='#aaaaaa',
                                                               opacity=0.6,
                                                               name=actorKey)
                self._redraw(which='subjectTrackerPosition')

        elif which in ('sampleBtns',):
            pointer = self.session.tools.pointer
            subjectTracker = self.session.tools.subjectTracker

            allowSampling = False
            if pointer is not None and subjectTracker is not None:
                if self._fidTblWdgt.currentCollectionItemKey is not None:
                    allowSampling = not any(self._positionsClient.getLatestTransf(key, None) is None for key in (pointer.trackerKey, subjectTracker.trackerKey))

            if self._sampleFiducialBtn.isEnabled() != allowSampling:
                self._sampleFiducialBtn.setEnabled(allowSampling)
            if self._sampleFiducialMultipleBtn.isEnabled() != allowSampling:
                self._sampleFiducialMultipleBtn.setEnabled(allowSampling)
            if self._sampleHeadPtsBtn.isEnabled() != allowSampling:
                self._sampleHeadPtsBtn.setEnabled(allowSampling)
            if self._newFidFromPointerBtn.isEnabled() != allowSampling:
                self._newFidFromPointerBtn.setEnabled(allowSampling)

        elif which in ('alignBtn',):
            self._alignToFiducialsBtn.setEnabled(self.session.subjectRegistration.hasMinimumSampledFiducials)

        elif which in ('initPointer', 'pointerPosition'):
            pointer = self.session.tools.pointer
            subjectTracker = self.session.tools.subjectTracker

            doShowPointer = self.session.subjectRegistration.trackerToMRITransf is not None \
                            and pointer is not None \
                            and (pointer.trackerSurf is not None or pointer.toolSurf is not None) \
                            and subjectTracker is not None

            if not doShowPointer:
                for toolOrTracker in ('tool', 'tracker'):
                    actorKey = 'pointer' + '_' + toolOrTracker
                    if actorKey in self._actors:
                        self._plotter.remove_actor(self._actors[actorKey])
                        self._actors.pop(actorKey)
                return

            if which == 'initPointer':
                for toolOrTracker in ('tool', 'tracker'):
                    actorKey = 'pointer' + '_' + toolOrTracker
                    tool = self.session.tools.pointer
                    actorSurf = getattr(tool, toolOrTracker + 'Surf')
                    meshColor = getattr(tool, toolOrTracker + 'Color')
                    if actorSurf is None:
                        if actorKey in self._actors:
                            self._plotter.remove_actor(self._actors[actorKey])
                            self._actors.pop(actorKey)
                        continue
                    self._actors[actorKey] = self._plotter.addMesh(mesh=actorSurf,
                                                                   color=meshColor,
                                                                   defaultMeshColor='#999999',
                                                                   opacity=0.6,
                                                                   name=actorKey)
                self._redraw(which='pointerPosition')

            elif which == 'pointerPosition':
                if not self.isVisible:
                    return  # don't do frequent updates when not even visible

                for toolOrTracker in ('tool', 'tracker'):
                    actorKey = 'pointer' + '_' + toolOrTracker

                    if actorKey not in self._actors:
                        # assume this was because we don't have enough info to show
                        continue

                    pointerToCameraTransf = self._positionsClient.getLatestTransf(pointer.trackerKey, None)
                    subjectTrackerToCameraTransf = self._positionsClient.getLatestTransf(subjectTracker.trackerKey, None)

                    if pointerToCameraTransf is None or subjectTrackerToCameraTransf is None:
                        # don't have valid info for determining pointer position relative to head tracker
                        if self._actors[actorKey].GetVisibility():
                            self._actors[actorKey].VisibilityOff()
                        continue

                    if not self._actors[actorKey].GetVisibility():
                        self._actors[actorKey].VisibilityOn()

                    if toolOrTracker == 'tool':
                        pointerStlToSubjectTrackerTransf = concatenateTransforms([
                            pointer.toolStlToToolTransf,
                            pointer.toolToTrackerTransf,
                            pointerToCameraTransf,
                            invertTransform(subjectTrackerToCameraTransf)
                        ])
                    elif toolOrTracker == 'tracker':
                        pointerStlToSubjectTrackerTransf = concatenateTransforms([
                            pointer.trackerStlToTrackerTransf,
                            pointerToCameraTransf,
                            invertTransform(subjectTrackerToCameraTransf)
                        ])
                    else:
                        raise NotImplementedError()

                    with self._plotter.allowNonblockingCalls():
                        setActorUserTransform(
                            self._actors[actorKey],
                            concatenateTransforms([
                                pointerStlToSubjectTrackerTransf,
                                self.session.subjectRegistration.trackerToMRITransf
                            ])
                        )
                        self._plotter.render()

            else:
                raise NotImplementedError()

        elif which == 'subjectTrackerPosition':
            if not self.isVisible:
                return  # don't do frequent updates when not even visible

            actorKey = 'subjectTracker'
            if actorKey not in self._actors:
                # subject tracker hasn't been initialized, maybe due to missing information
                return

            with self._plotter.allowNonblockingCalls():
                setActorUserTransform(
                    self._actors[actorKey],
                    self.session.subjectRegistration.trackerToMRITransf @ self.session.tools.subjectTracker.trackerStlToTrackerTransf
                )
                self._plotter.render()

        elif which == 'initPlannedFids':

            actorKey = 'plannedFids'

            labels = []
            coords = np.full((len(self.session.subjectRegistration.fiducials), 3), np.nan)
            for iFid, (label, fid) in enumerate(self.session.subjectRegistration.fiducials.items()):
                labels.append(label)
                if fid.plannedCoord is not None:
                    coords[iFid, :] = fid.plannedCoord

            if actorKey in self._actors:
                self._plotter.remove_actor(self._actors[actorKey])
                self._actors.pop(actorKey)

            self._actors[actorKey] = self._plotter.add_point_labels(
                name=actorKey,
                points=coords,
                labels=labels,
                point_color='blue',
                text_color='blue',
                point_size=12,
                shape=None,
                render_points_as_spheres=True,
                reset_camera=False,
                render=True
            )

        elif which == 'initSampledFids':

            actorKeys = ('sampledFids', 'repeatedSampledFids')

            sampledFidKeys = [key for key, fid in self.session.subjectRegistration.fiducials.items() if fid.sampledCoord is not None]
            doShowSampledFids = len(sampledFidKeys) > 0 \
                                and self.session.subjectRegistration.trackerToMRITransf is not None

            if not doShowSampledFids:
                # no sampled fiducials or necessary transform for plotting (yet)
                for actorKey in actorKeys:
                    if actorKey in self._actors:
                        self._plotter.remove_actor(actorKey)
                        self._actors.pop(actorKey)
                return

            labels = []
            coords = np.full((len(self.session.subjectRegistration.fiducials), 3), np.nan)
            repeatedCoords = np.zeros((0, 3))

            for iFid, (label, fid) in enumerate(self.session.subjectRegistration.fiducials.items()):
                labels.append(label)
                if fid.sampledCoord is not None:
                    coords[iFid, :] = fid.sampledCoord
                if fid.sampledCoords is not None and fid.sampledCoords.shape[0] > 1:
                    repeatedCoords = np.vstack((repeatedCoords, fid.sampledCoords))

            coords = applyTransform(self.session.subjectRegistration.trackerToMRITransf, coords)
            if repeatedCoords.shape[0] > 0:
                repeatedCoords = applyTransform(self.session.subjectRegistration.trackerToMRITransf, repeatedCoords)

            self._actors[actorKeys[0]] = self._plotter.add_point_labels(
                name=actorKeys[0],
                points=coords,
                labels=labels,
                point_color='green',
                text_color='green',
                point_size=15,
                shape=None,
                render_points_as_spheres=True,
                reset_camera=False,
                render=True
            )

            if repeatedCoords.shape[0] > 0:
                self._actors[actorKeys[1]] = self._plotter.add_points(
                    name=actorKeys[1],
                    points=repeatedCoords,
                    color='green',
                    opacity=0.7,
                    point_size=10,
                    render_points_as_spheres=True,
                    reset_camera=False,
                    render=True
                )
            else:
                actorKey = actorKeys[1]
                if actorKey in self._actors:
                    self._plotter.remove_actor(actorKey)
                    self._actors.pop(actorKey)

        elif which == 'initHeadPts':

            self._refineWithHeadpointsBtn.setEnabled(len(self.session.subjectRegistration.sampledHeadPoints) > 4 \
                                                     and self.session.subjectRegistration.trackerToMRITransf is not None)

            actorKeys = ['headPts', 'headPts_selected']

            doShowHeadPts = len(self.session.subjectRegistration.sampledHeadPoints) > 0 \
                            and self.session.subjectRegistration.trackerToMRITransf is not None

            if not doShowHeadPts:
                # no sampled head points or necessary transform (yet)
                for actorKey in actorKeys:
                    if actorKey in self._actors:
                        self._plotter.remove_actor(actorKey)
                        self._actors.pop(actorKey)
                return

            coords = applyTransform(self.session.subjectRegistration.trackerToMRITransf, np.asarray(self.session.subjectRegistration.sampledHeadPoints))

            # color selected points differently
            actorKey = actorKeys[1]
            selectedIndices = self._headPtsTblWdgt.selectedCollectionItemKeys

            if len(selectedIndices) > 0:
                selectedCoords = coords[selectedIndices, :]

                self._actors[actorKey] = self._plotter.add_points(
                    name=actorKey,
                    points=selectedCoords,
                    color='orange',
                    point_size=15,
                    render_points_as_spheres=True,
                    reset_camera=False,
                    render=False
                )

                unselectedIndices = list(range(coords.shape[0]))
                for index in reversed(sorted(selectedIndices)):
                    del unselectedIndices[index]
                coords = coords[unselectedIndices, :]
            else:
                if actorKey in self._actors:
                    self._plotter.remove_actor(actorKey)
                    self._actors.pop(actorKey)

            actorKey = actorKeys[0]
            if coords.shape[0] > 0:
                self._actors[actorKey] = self._plotter.add_points(
                    name=actorKey,
                    points=coords,
                    color='red',
                    point_size=10,
                    render_points_as_spheres=True,
                    reset_camera=False,
                    render=False
                )
            else:
                if actorKey in self._actors:
                    self._plotter.remove_actor(actorKey)
                    self._actors.pop(actorKey)

        else:
            raise NotImplementedError('Unexpected redraw key: {}'.format(which))

        with self._plotter.allowNonblockingCalls():
            self._plotter.render()



from __future__ import annotations

import asyncio
import attrs
import logging
import numpy as np
import pyvista as pv
import pytransform3d.rotations as ptr
from qtpy import QtWidgets, QtCore
import typing as tp
from typing import ClassVar

from NaviNIBS.Navigator.TargetingCoordinator import TargetingCoordinator
from .ViewLayers import ViewLayer, PlotViewLayer, LegendEntry
from .ViewLayers.MeshSurfaceLayer import HeadMeshSurfaceLayer, ToolMeshSurfaceLayer
from .ViewLayers.OrientationsLayers import SampleOrientationsLayer, TargetOrientationsLayer
from .ViewLayers.SampleMetadataOrientationsLayer import SampleMetadataOrientationsLayer, SampleMetadataInterpolatedSurfaceLayer
from .ViewLayers.TargetingCrosshairsLayer import TargetingCoilCrosshairsLayer, TargetingTargetCrosshairsLayer
from .ViewLayers.TargetingAngleErrorLayer import TargetingAngleErrorLayer
from .ViewLayers.TargetingPointLayer import TargetingCoilPointsLayer, TargetingTargetPointsLayer
from. ViewLayers.TargetingErrorLineLayer import TargetingErrorLineLayer

from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.Transforms import applyTransform, composeTransform, concatenateTransforms
from NaviNIBS.util.GUI.Dock import Dock
from NaviNIBS.util.GUI.QueuedRedrawMixin import QueuedRedrawMixin
from NaviNIBS.util.numpy import array_equalish
from NaviNIBS.util.pyvista import DefaultPrimaryLayeredPlotter, RemotePlotterProxy

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


Transform = np.ndarray


@attrs.define
class NavigationView(QueuedRedrawMixin):
    _key: str
    _type: ClassVar[str]
    _coordinator: TargetingCoordinator

    _dockKeyPrefix: str = ''
    _title: str | None = None

    _dock: Dock = attrs.field(init=False)
    _wdgt: QtWidgets.QWidget = attrs.field(init=False)
    _contextMenu: QtWidgets.QMenu = attrs.field(init=False)

    _layers: tp.Dict[str, ViewLayer] = attrs.field(init=False, factory=dict)
    _layerLibrary: tp.Dict[str, tp.Callable[..., ViewLayer]] = attrs.field(init=False, factory=dict, repr=False)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._dock = Dock(
            name=self._dockKeyPrefix + self._key,
            title=self._title,
            affinities=[self._dockKeyPrefix])

        self._wdgt = QtWidgets.QWidget()
        self._wdgt.setContentsMargins(0, 0, 0, 0)
        self._dock.addWidget(self._wdgt)

        self._wdgt.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

        # TODO: add context menu to be able to change view type with right click on title bar (?)

    def addLayer(self, key: str, type: str, **kwargs):
        raise NotImplementedError()  # should be implemented by subclass

    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str, ...]] = None):
        #logger.debug(f'redraw {which}')

        super()._redraw(which=which)

        if which is None:
            which = 'all'
            self._redraw(which=which)
            return

        if not isinstance(which, str):
            for subWhich in which:
                self._redraw(which=subWhich)
            return

        # subclass should handle the rest

    @property
    def key(self):
        return self._key

    @property
    def type(self):
        return self._type

    @property
    def dock(self):
        return self._dock

    @property
    def wdgt(self):
        return self._wdgt


@attrs.define
class SinglePlotterNavigationView(NavigationView):
    _plotter: DefaultPrimaryLayeredPlotter = attrs.field(init=False, repr=False)
    _plotInSpace: str = 'MRI'
    _alignCameraTo: str | None = None
    _alignCameraOffset: tuple[float, float, float] | None = None
    _cameraDist: float = 100
    """
    None to use default camera perspective; 'target' to align camera space to target space, etc.
    Can also specify 'target-Y' to look at from along typical coil handle axis, or 'Target+X' to look at from other direction
    Can also specify something like 'tool-<toolKey>+X' to look at from along tool X axis
    """
    _doParallelProjection: bool = False

    _doShowLegend: bool = False

    _layers: tp.Dict[str, PlotViewLayer] = attrs.field(init=False, factory=dict)
    _layerLibrary: tp.Dict[str, tp.Callable[..., PlotViewLayer]] = attrs.field(init=False, factory=dict, repr=False)
    _lastAlignedToolPose: Transform | None = attrs.field(init=False, default=None, repr=False)

    finishedAsyncInit: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._wdgt.setLayout(QtWidgets.QVBoxLayout())
        self._wdgt.layout().setContentsMargins(0, 0, 0, 0)

        self._plotter = DefaultPrimaryLayeredPlotter()
        self._wdgt.layout().addWidget(self._plotter)

        self._layerLibrary = {}
        for cls in (
                    HeadMeshSurfaceLayer,
                    ToolMeshSurfaceLayer,
                    SampleOrientationsLayer,
                    SampleMetadataOrientationsLayer,
                    SampleMetadataInterpolatedSurfaceLayer,
                    TargetOrientationsLayer,
                    TargetingTargetCrosshairsLayer,
                    TargetingCoilCrosshairsLayer,
                    TargetingTargetPointsLayer,
                    TargetingCoilPointsLayer,
                    TargetingErrorLineLayer,
                    TargetingAngleErrorLayer,
                    ):
            self._layerLibrary[cls.type] = cls

        asyncio.create_task(asyncTryAndLogExceptionOnError(self.__finishInitialization_async))

    async def __finishInitialization_async(self):
        # call _finishInitialization_async() here to allow extending in subclass while still setting event only after fully initialized
        await self._finishInitialization_async()
        self.finishedAsyncInit.set()

    async def _finishInitialization_async(self):
        if isinstance(self._plotter, RemotePlotterProxy):
            await self._plotter.isReadyEvent.wait()

        self._coordinator.sigCurrentTargetChanged.connect(self._onCurrentTargetChanged)
        self._coordinator.sigCurrentCoilPositionChanged.connect(self._onCurrentCoilPositionChanged)
        self._coordinator.positionsClient.sigLatestPositionsChanged.connect(self._onLatestPositionsChanged)

        if self._doParallelProjection:
            self._plotter.camera.enable_parallel_projection()

        self._redraw('all')

    @property
    def plotter(self):
        return self._plotter

    def _onCurrentTargetChanged(self):
        if self._alignCameraTo.startswith('target'):
            self._queueRedraw(which='camera')

    def _onCurrentCoilPositionChanged(self):
        if self._alignCameraTo.startswith('coil'):
            self._queueRedraw(which='camera')

    def _onLatestPositionsChanged(self):
        if self._alignCameraTo.startswith('tool-'):
            if self._alignCameraTo[-2:] in ('+X', '-X', '+Y', '-Y', '+Z', '-Z'):
                toolKey = self._alignCameraTo[len('tool-'):-2]
            else:
                toolKey = self._alignCameraTo[len('tool-'):]

            tool = self._coordinator.session.tools[toolKey]
            trackerKey = tool.trackerKey

            newTrackerPose = self._coordinator.positionsClient.getLatestTransf(trackerKey, None)
            if newTrackerPose is None:
                newToolPose = None
            else:
                if tool.toolToTrackerTransf is None:
                    newToolPose = None
                else:
                    # TODO: do some extra caching to not recalculate this with every position change
                    # (since the tool we align to may not actually be changing frequently)
                    # but then would also need to connect to toolTrackerTransf changed signal.
                    newToolPose = concatenateTransforms((tool.toolToTrackerTransf, newTrackerPose))

            if not array_equalish(newToolPose, self._lastAlignedToolPose):
                self._lastAlignedToolPose = newToolPose  # not actually aligned yet due to queueing, but should be okay unless there
                # are rapid bursts of switching between a small set of discrete poses
                self._queueRedraw(which='camera')

    @staticmethod
    def _getExtraRotationForToAlignCamera(rotSuffix: str) -> np.ndarray:
        # TODO: double check angles and signs
        match rotSuffix:
            case '':
                extraRot = np.eye(3)
            case '+X':
                extraRot = ptr.active_matrix_from_intrinsic_euler_yzy([np.pi / 2, np.pi/2, 0])
            case '-X':
                extraRot = ptr.active_matrix_from_intrinsic_euler_yzy([-np.pi / 2, -np.pi/2, 0])
            case '+Y':
                extraRot = ptr.active_matrix_from_extrinsic_euler_xyz([-np.pi / 2, 0, np.pi])
            case '-Y':
                extraRot = ptr.active_matrix_from_extrinsic_euler_xyz([np.pi / 2, 0, 0])
            case '+Z':
                extraRot = np.eye(3)
            case '-Z':
                extraRot = ptr.active_matrix_from_extrinsic_euler_xyz([np.pi, 0, 0])
            case _:
                raise NotImplementedError
        return extraRot

    def _alignCamera(self):
        class NoValidCameraPoseAvailable(Exception):
            pass

        try:
            cameraPts = np.asarray([[0, 0, 0], [0, 0, self._cameraDist], [0, 1, 0]])  # focal point, position, and up respectively

            if self._alignCameraTo is None:
                with self._plotter.allowNonblockingCalls:
                    self._plotter.reset_camera()
                    self._plotter.render()

            elif self._alignCameraTo.startswith('target'):
                extraRot = self._getExtraRotationForToAlignCamera(self._alignCameraTo[len('target'):])
                extraTransf = composeTransform(extraRot)

                if self._alignCameraOffset is not None:
                    extraTransf[0:3, 3] = np.asarray(self._alignCameraOffset)

                if self._plotInSpace == 'MRI':
                    if self._coordinator.currentTarget is not None and self._coordinator.currentTarget.coilToMRITransf is not None:
                        cameraPts = applyTransform(self._coordinator.currentTarget.coilToMRITransf @ extraTransf, cameraPts,
                                                   doCheck=False)
                    else:
                        raise NoValidCameraPoseAvailable()
                else:
                    raise NotImplementedError()

            elif self._alignCameraTo.startswith('coil'):
                extraRot = self._getExtraRotationForToAlignCamera(self._alignCameraTo[len('coil'):])
                extraTransf = composeTransform(extraRot)

                if self._alignCameraOffset is not None:
                    extraTransf[0:3, 3] = np.asarray(self._alignCameraOffset)

                if self._plotInSpace == 'MRI':
                    if self._coordinator.currentCoilToMRITransform is not None:
                        cameraPts = applyTransform(self._coordinator.currentCoilToMRITransform @ extraTransf, cameraPts, doCheck=False)
                    else:
                        raise NoValidCameraPoseAvailable()
                else:
                    raise NotImplementedError()

            elif self._alignCameraTo.startswith('tool-'):
                if self._alignCameraTo[-2:] in ('+X', '-X', '+Y', '-Y', '+Z', '-Z'):
                    extraRot = self._getExtraRotationForToAlignCamera(self._alignCameraTo[-2:])
                    extraTransf = composeTransform(extraRot)
                    toolKey = self._alignCameraTo[len('tool-'):-2]
                else:
                    extraTransf = np.eye(4)
                    toolKey = self._alignCameraTo[len('tool-'):]

                if self._alignCameraOffset is not None:
                    extraTransf[0:3, 3] = np.asarray(self._alignCameraOffset)

                trackerKey = self._coordinator.session.tools[toolKey].trackerKey
                trackerToWorldTransf = self._coordinator.positionsClient.getLatestTransf(trackerKey, None)

                if trackerToWorldTransf is None:
                    # missing information
                    raise NoValidCameraPoseAvailable

                if self._plotInSpace == 'MRI':
                    raise NotImplementedError  # TODO

                elif self._plotInSpace == 'World':
                    toolToTrackerTransf = self._coordinator.session.tools[toolKey].toolToTrackerTransf

                    if toolToTrackerTransf is None:
                        # missing information
                        raise NoValidCameraPoseAvailable

                    cameraPts = applyTransform(trackerToWorldTransf @ toolToTrackerTransf @ extraTransf, cameraPts, doCheck=False)

            else:
                 raise NotImplementedError()

        except NoValidCameraPoseAvailable:
            #logger.debug('No camera pose available')
            if False:
                with self._plotter.allowNonblockingCalls():
                    self._plotter.reset_camera()
                    self._plotter.reset_camera_clipping_range()
                    self._plotter.render()
                return  # TODO: change display to indicate view is out-of-date / invalid
            else:
                if False:
                    # continue plotting, but with generic position
                    cameraPts = np.asarray([[0, 0, 0], [0, 0, self._cameraDist * 3], [0, 1, 0]])  # focal point, position, and up respectively
                else:
                    # hide plot until we have necessary info
                    self._wdgt.setVisible(False)
                    return

        with self._plotter.allowNonblockingCalls():
            with self._plotter.renderingPaused():
                self._plotter.camera.focal_point = cameraPts[0, :]
                self._plotter.camera.position = cameraPts[1, :]
                self._plotter.camera.up = cameraPts[2, :] - cameraPts[1, :]
                if True:
                    # force fixed zoom in parallel camera views
                    self.plotter.camera.parallel_scale = self._cameraDist / 2

                if False:
                    # force fixed zoom in perspective camera views
                    self.plotter.camera.view_angle = 60.

                if self._alignCameraTo[-1] in 'XY' and False:
                    # orthogonal view, clip camera
                    self._plotter.camera.clipping_range = (self._cameraDist-0.1, self._cameraDist+0.1)
                else:
                    self._plotter.reset_camera_clipping_range()
            self._plotter.render()

        if not self._wdgt.isVisible():
            self._wdgt.setVisible(True)

    def addLayer(self, type: str, key: str, layeredPlotterKey: tp.Optional[str] = None,
                 plotterLayer: tp.Optional[int] = None, **kwargs):
        """
        Note: layer is added aynchronously later, to support delayed plotter initialization
        """
        asyncio.create_task(asyncTryAndLogExceptionOnError(self._addLayer_async,
                                                           type=type,
                                                           key=key,
                                                           layeredPlotterKey=layeredPlotterKey,
                                                           plotterLayer=plotterLayer,
                                                           **kwargs
                                                           ))

    async def _addLayer_async(self, type: str, key: str, layeredPlotterKey: tp.Optional[str] = None,
                              plotterLayer: tp.Optional[int] = None, **kwargs):

        if isinstance(self._plotter, RemotePlotterProxy):
            await self._plotter.isReadyEvent.wait()

        cls = self._layerLibrary[type]
        assert key not in self._layers
        if layeredPlotterKey is None:
            plotter = self._plotter
            if plotterLayer is not None:
                plotter.setLayer(plotterLayer)
        else:
            if layeredPlotterKey in self._plotter.secondaryPlotters:
                plotter = self._plotter.secondaryPlotters[layeredPlotterKey]
                if plotterLayer is not None:
                    assert plotter.rendererLayer == plotterLayer
            else:
                plotter = self._plotter.addLayeredPlotter(key=layeredPlotterKey, layer=plotterLayer)
                logger.debug(f'Added renderer layer {layeredPlotterKey} #{plotter.rendererLayer}')

        self._layers[key] = cls(key=key, **kwargs,
                                coordinator=self._coordinator,
                                plotter=plotter,
                                plotInSpace=self._plotInSpace)

        plotter.render()

    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str]] = None):

        if isinstance(self._plotter, RemotePlotterProxy) and not self._plotter.isReadyEvent.is_set():
            # plotter not yet ready
            return

        super()._redraw(which=which)

        if not isinstance(which, str):
            # assume parent call above triggered appropriate redraws
            return

        if which == 'all':
            which = ['layers', 'camera']
            self._redraw(which=which)
            return

        if which == 'camera':
            self._alignCamera()

        elif which == 'layers':
            for layer in self._layers.values():
                layer._queueRedraw(which='all')

        else:
            raise NotImplementedError


@attrs.define
class TargetingCrosshairsView(SinglePlotterNavigationView):
    _type: ClassVar[str] = 'TargetingCrosshairs'
    _alignCameraTo: str = 'target'
    _doShowSkinSurf: bool = False
    _doShowHandleAngleError: bool = False
    _doShowTargetTangentialAngleError: bool = False
    _doShowScalpTangentialAngleError: bool = False

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        plotLayer = 0

        if self._doShowSkinSurf:
            self.addLayer(type='HeadMeshSurface', key='Skin', surfKey='skinSimpleSurf',
                          color='#c9c5c2',
                          layeredPlotterKey='SkinMesh',
                          plotterLayer=plotLayer)

            plotLayer += 1

            #self._plotter.secondaryPlotters['SkinMesh'].enable_depth_peeling(2)


        if False and self._alignCameraTo == 'target':
            self.addLayer(type='SampleMetadataInterpolatedSurface',
                          key='Brain',
                          surfKey='gmSurf',
                          metadataKey='Vpp_dBmV',
                          colorbarLabel='Vpp (dBmV)',
                          scalarAnnotations={
                              20*np.log10(50e-3): '50 uV',
                              20*np.log10(1): '1 mV',
                          },
                          relevantSampleDepth='intersection',
                          kernelRadius=8,
                          layeredPlotterKey='Brain',
                          plotterLayer=plotLayer)
        else:
            self.addLayer(type='HeadMeshSurface', key='Brain', surfKey='gmSurf',
                          layeredPlotterKey='Brain',
                          plotterLayer=plotLayer)

        plotLayer += 1

        if False and self._alignCameraTo == 'target':
            self.addLayer(type='SampleMetadataInterpolatedSurface',
                          key='ScalpVpps',
                          surfKey='csfSurf',
                          opacity=0.5,
                          meshOpacityOutsideInterpolatedRegion=0.,
                          metadataKey='Vpp_dBmV',
                          colorbarLabel='Vpp (dBmV)',
                          scalarAnnotations={
                              20 * np.log10(50e-3): '50 uV',
                              20 * np.log10(1): '1 mV',
                          },
                          relevantSampleDepth='intersection',
                          layeredPlotterKey='ScalpVpps',
                          plotterLayer=plotLayer)
            plotLayer += 1

        if True:
            self.addLayer(type='SampleOrientations', key='Samples',
                          layeredPlotterKey='Orientations',
                          plotterLayer=plotLayer)
        elif False:
            self.addLayer(type='SampleMetadataOrientations',
                          key='SampleVpps',
                          layeredPlotterKey='Orientations',
                          plotterLayer=plotLayer,
                          metadataKey='Vpp',
                          metadataScaleFactor=1.e6,
                          colorbarLabel='Vpp (uV)',
                          lineWidth=6.)
        else:
            self.addLayer(type='SampleMetadataOrientations',
                          key='SampleVpps_dBmV',
                          layeredPlotterKey='Orientations',
                          plotterLayer=plotLayer,
                          metadataKey='Vpp_dBmV',
                          colorbarLabel='Vpp (dBmV)',
                          lineWidth=6.)

        self.addLayer(type='TargetOrientations', key='Targets', layeredPlotterKey='Orientations')
        self.addLayer(type='TargetingTargetPoints', key='TargetPoints', layeredPlotterKey='Orientations')
        self.addLayer(type='TargetingCoilPoints', key='CoilPoints', layeredPlotterKey='Orientations')

        plotLayer += 1

        self.addLayer(type='TargetingTargetCrosshairs', key='Target',
                      layeredPlotterKey='Crosshairs',
                      plotterLayer=plotLayer)
        self.addLayer(type='TargetingCoilCrosshairs', key='Coil', layeredPlotterKey='Crosshairs')
        plotLayer += 1

        self.addLayer(type='TargetingErrorLine', key='TargetError', targetDepth='target', coilDepth='target',
                      layeredPlotterKey='TargetingError',
                      plotterLayer=plotLayer)
        self.addLayer(type='TargetingErrorLine', key='CoilError', targetDepth='coil', coilDepth='coil',
                      layeredPlotterKey='TargetingError')

        if self._doShowHandleAngleError:
            self.addLayer(type='TargetingAngleError', key='HandleAngleError',
                          angleMetric='Horiz angle error',
                          layeredPlotterKey='TargetingError')

        if self._doShowTargetTangentialAngleError:
            plotOn = self._alignCameraTo[:-2]
            if plotOn == 'coil':
                multiplier = -4.
            elif plotOn == 'target':
                multiplier = -4.
            else:
                raise NotImplementedError
            if self._alignCameraTo.endswith('X'):
                xyDims = (1, 2)
                angleMetric = f'Depth {plotOn} Y angle error'
            elif self._alignCameraTo.endswith('Y'):
                xyDims = (0, 2)
                plotOn = self._alignCameraTo[:-2]
                angleMetric = f'Depth {plotOn} X angle error'
            else:
                raise NotImplementedError
            self.addLayer(type='TargetingAngleError', key='TargetTangentialAngleError',
                          plotOnTargetOrCoil=plotOn,
                          angleMetric=angleMetric,
                          angleOffset=np.pi/2,
                          multiplier=multiplier,
                          radius=15,
                          xyDims=xyDims,
                          layeredPlotterKey='TargetingError')

        if self._doShowScalpTangentialAngleError:
            plotOn = self._alignCameraTo[:-2]
            if plotOn == 'coil':
                multiplier = -4.
            elif plotOn == 'target':
                multiplier = -4.
            else:
                raise NotImplementedError
            if self._alignCameraTo.endswith('X'):
                xyDims = (1, 2)
                angleMetric = f'Normal {plotOn} Y angle error'
            elif self._alignCameraTo.endswith('Y'):
                xyDims = (0, 2)
                plotOn = self._alignCameraTo[:-2]
                angleMetric = f'Normal {plotOn} X angle error'
            else:
                raise NotImplementedError
            self.addLayer(type='TargetingAngleError', key='ScalpTangentialAngleError',
                          color='#9fc58b',
                          plotOnTargetOrCoil=plotOn,
                          angleMetric=angleMetric,
                          angleOffset=np.pi/2,
                          multiplier=multiplier,
                          radius=12,
                          xyDims=xyDims,
                          layeredPlotterKey='TargetingError')

            plotLayer += 1

    async def _finishInitialization_async(self):
        await super()._finishInitialization_async()

        #self._plotter.enable_depth_peeling(2)



@attrs.define
class TargetingOverMeshSurfacesView(SinglePlotterNavigationView):
    _type: ClassVar[str] = 'TargetingOverMeshSurfaces'
    _viewRelativeTo: str = 'perspective'

    # TODO


@attrs.define
class TargetingSliceView(SinglePlotterNavigationView):
    _type: ClassVar[str] = 'TargetingSlice'
    _viewRelativeTo: str = 'coil'
    _normal: tp.Union[str, np.ndarray] = 'x'  # axis in space defined by self.viewRelativeTo

    _backgroundColor: str = '#000000'

    # TODO



import json
import numpy as np
from pytransform3d.transformations import check_transform
from qtpy import QtWidgets, QtCore, QtGui
import re
import typing as tp

from NaviNIBS.util.numpy import array_equalish
from NaviNIBS.util.Transforms import transformToString, stringToTransform


invalidRegex = re.compile(r'(\[\[\[)|([a-zA-Z]+)|(\d+ +[\d.]+)|(]]])|(,,)|(\.\.)|(] +\[)')


class TransformValidator(QtGui.QValidator):
    def validate(self, inputStr: str, pos: int) -> tp.Tuple[QtGui.QValidator.State, str, int]:
        try:
            stringToTransform(inputStr)
        except ValueError as e:
            if invalidRegex.search(inputStr):
                return self.State.Invalid, inputStr, pos
            else:
                # TODO: add detection other clearly invalid cases
                return self.State.Intermediate, inputStr, pos
        else:
            return self.State.Acceptable, inputStr, pos


class OptionalTransformValidator(TransformValidator):
    def validate(self, inputStr: str, pos: int) -> tp.Tuple[QtGui.QValidator.State, str, int]:
        if len(inputStr.strip()) == 0:
            return self.State.Acceptable, inputStr, pos
        else:
            return super().validate(inputStr=inputStr, pos=pos)





import asyncio
import attrs
import logging
import pylsl as lsl
import qtawesome as qta
from qtpy import QtWidgets, QtCore, QtGui
import typing as tp

from NaviNIBS.util.lsl.LSLStreamResolver import ThreadedLSLStreamResolver
from NaviNIBS.util.GUI.Icons import getIcon
from NaviNIBS.util.Signaler import Signal


logger = logging.getLogger(__name__)


@attrs.define
class LSLStreamSelector:
    _wdgt: QtWidgets.QWidget = attrs.field(factory=QtWidgets.QWidget)
    _streamKeys: list[str] = attrs.field(factory=list)
    _selectedStreamKey: tp.Optional[str] = None

    _resolver: ThreadedLSLStreamResolver = attrs.field(init=False)
    _comboBox: QtWidgets.QComboBox = attrs.field(init=False)

    _icon_available: QtGui.QIcon = attrs.field(factory=lambda: getIcon('mdi6.eye'))
    _icon_unavailable: QtGui.QIcon = attrs.field(factory=lambda: qta.icon('mdi6.eye-off', color='gray'))

    sigSelectedStreamKeyChanged: Signal = attrs.field(init=False, factory=lambda: Signal((str,)))
    sigSelectedStreamAvailabilityChanged: Signal = attrs.field(init=False, factory=Signal)  # (note: not emitted when key changes to a stream with different availability)

    def __attrs_post_init__(self):
        self._resolver = ThreadedLSLStreamResolver()
        self._resolver.sigStreamDetected.connect(self._onStreamDetected)
        self._resolver.sigStreamLost.connect(self._onStreamLost)

        self._wdgt.setLayout(QtWidgets.QVBoxLayout())

        self._comboBox = QtWidgets.QComboBox()
        self._comboBox.setEditable(True)
        self._comboBox.setInsertPolicy(self._comboBox.InsertPolicy.InsertAtBottom)
        self._comboBox.currentIndexChanged.connect(self._onComboBoxIndexChanged)
        self._wdgt.layout().addWidget(self._comboBox)

        if self._selectedStreamKey is not None and self._selectedStreamKey not in self._streamKeys:
            self._streamKeys.append(self._selectedStreamKey)

        for key in self._resolver.availableStreams:
            if key not in self._streamKeys:
                self._streamKeys.append(key)

        self._updateComboBox()

    @property
    def wdgt(self):
        return self._wdgt

    @property
    def selectedStreamKey(self):
        return self._selectedStreamKey

    @selectedStreamKey.setter
    def selectedStreamKey(self, key: tp.Optional[str]):
        if self._selectedStreamKey == key:
            return
        logger.info(f'Selected stream key changed: {key}')
        if key not in self._streamKeys:
            self._streamKeys.append(key)
        self._selectedStreamKey = key
        self._updateComboBox()

        self.sigSelectedStreamKeyChanged.emit(key)

    @property
    def selectedStreamIsAvailable(self):
        return self._selectedStreamKey is not None and self._selectedStreamKey in self._resolver.availableStreams

    @property
    def selectedStreamInfo(self) -> tp.Optional[lsl.StreamInfo]:
        return self._resolver.availableStreams.get(self._selectedStreamKey, None)

    def markStreamAsLost(self, streamKey: str):
        self._resolver.markStreamAsLost(streamKey=streamKey)

    def _updateComboBox(self):
        for key in self._streamKeys:
            isAvailable = key in self._resolver.availableStreams
            icon = self._icon_available if isAvailable else self._icon_unavailable
            index = self._comboBox.findText(key)
            if index == -1:
                # key not already in combo box
                self._comboBox.addItem(icon, key)
            else:
                # update available icon
                self._comboBox.setItemIcon(index, icon)
        # note: we don't handle case where a key is removed from streamKeys but is still in combo box (for now)

        if self._selectedStreamKey is None:
            if self._comboBox.currentIndex() != -1:
                self._comboBox.setCurrentIndex(-1)
        else:
            assert self._selectedStreamKey in self._streamKeys
            index = self._comboBox.findText(self._selectedStreamKey)
            if self._comboBox.currentIndex() != index:
                self._comboBox.setCurrentIndex(index)

    def _onStreamDetected(self, key: str, info: lsl.StreamInfo):
        if key not in self._streamKeys:
            self._streamKeys.append(key)
        self._updateComboBox()
        if key == self._selectedStreamKey:
            self.sigSelectedStreamAvailabilityChanged.emit()

    def _onStreamLost(self, key: str, info: lsl.StreamInfo):
        self._updateComboBox()
        if key == self._selectedStreamKey:
            self.sigSelectedStreamAvailabilityChanged.emit()

    def _onComboBoxIndexChanged(self, index: int):
        newStreamKey = self._comboBox.itemText(index)
        logger.debug(f'Combo box changed selected stream key from {self._selectedStreamKey} to {newStreamKey}')
        self.selectedStreamKey = newStreamKey
